# pylint: disable=wrong-import-position
"""
Tracker3 plugin using gir1.2-tracker-3.0
"""

from __future__ import annotations

__kupfer_name__ = _("Tracker3")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ("Tracker3Fulltext",)
__kupfer_contents__ = ()
__kupfer_actions__ = ("TrackerSearch",)
__description__ = _("Tracker3 search integration")
__version__ = "2023-04-30"
__author__ = "KB"

import typing as ty

import gi

try:
    gi.require_version("Tracker", "3.0")
except ValueError as exc:
    raise ImportError(
        f"{exc}: missing GIRepository Tracker library (gir1.2-tracker-3.0)"
    ) from exc

# pylint: disable=no-name-in-module
from gi.repository import Gio, Tracker

from kupfer.obj import (
    Action,
    FileLeaf,
    Source,
    TextLeaf,
    TextSource,
)

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class TrackerSearch(Action):
    def __init__(self):
        Action.__init__(self, _("Search in Tracker3"))

    def is_factory(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        return QuerySource(leaf.object.strip())

    def get_description(self):
        return _("Search in Tracker3")

    def get_icon_name(self):
        return "system-search"

    def item_types(self):
        yield TextLeaf


_QUERY = """
SELECT ?uri max(?rank) as ?rank WHERE {
    {
        SELECT (?s AS ?uri) (fts:rank(?s) AS ?rank) {
            GRAPH tracker:FileSystem {
                ?s a nfo:FileDataObject ;
                fts:match "%(term)s" ;
                nie:dataSource ?ds .
            }
        }
    }
    UNION {
        SELECT ?uri (fts:rank(?uri) AS ?rank) {
            GRAPH ?g {
                ?s a nie:InformationElement ;
                fts:match "%(term)s" ;
                nie:isStoredAs ?uri .
            }
            GRAPH tracker:FileSystem {
                ?uri nie:dataSource ?ds .
            }
        }
    }
}
GROUP BY ?uri
ORDER BY ?rank
OFFSET 0
LIMIT %(limit)s
"""


def _query(term: str, limit: int) -> ty.Iterator[FileLeaf]:
    miner_fs = Tracker.SparqlConnection.bus_new(
        "org.freedesktop.Tracker3.Miner.Files", None, None
    )
    term = Tracker.sparql_escape_string(term)
    query = _QUERY % {"term": term, "limit": limit}
    if cursor := miner_fs.query(query, None):
        while cursor.next(None):
            if uri := cursor.get_string(0)[0]:
                path = Gio.File.new_for_uri(uri).get_path()
                yield FileLeaf(path)


class QuerySource(Source):
    def __init__(self, query: str, limit: int = 100) -> None:
        Source.__init__(self, name=_('Results for "%s"') % query)
        self.query = query
        self.limit = limit

    def repr_key(self):
        return self.query

    def get_items(self):
        return tuple(_query(self.query, self.limit))


class Tracker3Fulltext(TextSource):
    def __init__(self):
        TextSource.__init__(self, name=_("Tracker3 Full Text Search"))

    def get_description(self):
        return _("Use '?' prefix to get full text results")

    def get_text_items(self, text):
        if text.startswith("?") and (term := text.lstrip("? ")):
            return tuple(_query(term, 25))

        return ()

    def provides(self):
        yield FileLeaf

    def get_rank(self):
        return 80



from __future__ import annotations

from gi.repository import Gtk

from kupfer import config, version


class _CredentialsDialogController:
    def __init__(
        self,
        username: str | None,
        password: str | None,
        infotext: str | None = None,
    ):
        """Load ui from data file"""
        builder = Gtk.Builder()
        builder.set_translation_domain(version.PACKAGE_NAME)
        ui_file = config.get_data_file("credentials_dialog.ui")
        builder.add_from_file(ui_file)
        builder.connect_signals(self)  # pylint: disable=no-member

        self.window = builder.get_object("credentials_dialog")
        self.entry_user = builder.get_object("entry_username")
        self.entry_pass = builder.get_object("entry_password")
        if infotext:
            hbox_information = builder.get_object("hbox_information")
            label_information = builder.get_object("label_information")
            label_information.set_text(infotext)
            hbox_information.show()

        self.entry_user.set_text(username or "")
        self.entry_pass.set_text(password or "")

    def on_button_ok_clicked(self, widget: Gtk.Widget) -> None:
        self.window.response(Gtk.ResponseType.ACCEPT)
        self.window.hide()

    def on_button_cancel_clicked(self, widget: Gtk.Widget) -> None:
        self.window.response(Gtk.ResponseType.CANCEL)
        self.window.hide()

    def show(self) -> bool:
        return self.window.run()  # type: ignore

    @property
    def username(self) -> str:
        return self.entry_user.get_text()  # type:ignore

    @property
    def password(self) -> str:
        return self.entry_pass.get_text()  # type:ignore


def ask_user_credentials(
    user: str | None = None,
    password: str | None = None,
    infotext: str | None = None,
) -> tuple[str, str] | None:
    """Ask user for username and password.

    @user, @password - initial values
    @return:
    (user, password) when user press "change"
    None when user press "cancel" button"""
    dialog = _CredentialsDialogController(user, password, infotext)
    result = None
    if dialog.show() == Gtk.ResponseType.ACCEPT:
        result = dialog.username, dialog.password

    return result



from __future__ import annotations

import contextlib
import os
import typing as ty
from dataclasses import dataclass

from gi.repository import Gdk, Gtk

from kupfer.support import pretty
from kupfer.ui import keybindings


def try_close_unused_displays(screen: Gdk.Screen) -> None:
    """Try to close inactive displays.

    `screen` is current GdkScreen.

    Take all GtkWindow that are hidden, and move to the
    current screen. If no windows remain then we close
    the display, but we never close the default display.
    """

    skip_displays = [
        d.get_name()
        for d in (screen.get_display(), Gdk.Display.get_default())
        if d
    ]
    if not skip_displays:
        pretty.print_error(__name__, "empty skip_displays")
        return

    dmgr = Gdk.DisplayManager.get()
    for disp in dmgr.list_displays():
        dname = disp.get_name()
        if dname in skip_displays:
            continue

        pretty.print_debug(__name__, "Trying to close", dname)
        open_windows = 0
        for window in Gtk.Window.list_toplevels():
            # find windows on @disp
            if window.get_screen().get_display().get_name() != dname:
                continue

            if not window.get_property("visible"):
                pretty.print_debug(
                    __name__, "Moving window", window.get_name()
                )
                pretty.print_debug(__name__, "Moving", window.get_title())
                window.set_screen(screen)
            else:
                pretty.print_debug(__name__, "Open window blocks close")
                open_windows += 1

        if not open_windows:
            pretty.print_debug(__name__, "Closing display", dname)
            disp.close()


class GUIEnvironmentContext:
    """Context object for action execution in the current GUI context"""

    def __init__(
        self, timestamp: int, screen: Gdk.Screen | None = None
    ) -> None:
        self._timestamp = timestamp
        self._screen: Gdk.Screen = screen or Gdk.Screen.get_default()

    def __repr__(self):
        return (
            f"<{type(self).__name__} "
            f"time={self._timestamp!r} screen={self._screen!r}>"
        )

    @classmethod
    def ensure_display_open(cls, display_name: str | None) -> Gdk.Display:
        """Get GdkDisplay for `display_name` or return default if `display_name`
        is None or invalid/unavailable."""
        # NOTE: display parameter was not supported; added (k)
        if disp := Gdk.Display.open(display_name):
            return disp

        return Gdk.DisplayManager.get().get_default_display()

    def get_timestamp(self) -> int:
        return self._timestamp

    def get_startup_notification_id(self) -> str:
        """Get new notification id."""
        return _make_startup_notification_id(self.get_timestamp())

    def get_display(self) -> str:
        """Return the display name to show new windows on."""
        return ty.cast("str", self._screen.make_display_name())

    def get_screen(self) -> Gdk.Screen:
        return self._screen

    def present_window(self, window: Gtk.Window) -> None:
        """Show and present `window` on the current workspace, screen & display
        as appropriate."""
        window.set_screen(self.get_screen())
        window.present_with_time(self.get_timestamp())


def gui_context_from_widget(
    timestamp: int, widget: Gtk.Widget
) -> GUIEnvironmentContext:
    return GUIEnvironmentContext(timestamp, widget.get_screen())


def gui_context_from_timestamp(timestamp: int) -> GUIEnvironmentContext:
    return GUIEnvironmentContext(timestamp, None)


def gui_context_from_keyevent(
    timestamp: int, display: str
) -> GUIEnvironmentContext:
    new_display = GUIEnvironmentContext.ensure_display_open(display)
    screen, *_dummy = new_display.get_pointer()
    return GUIEnvironmentContext(timestamp, screen)


@dataclass
class _InternalData:
    seq: int = 0
    current_event_time: int = 0

    def inc_seq(self) -> None:
        self.seq = self.seq + 1


_internal_data = _InternalData()


def _make_startup_notification_id(time: int) -> str:
    _internal_data.inc_seq()
    return f"kupfer-%{os.getpid()}-{_internal_data.seq}_TIME{time}"


def current_event_time() -> int:
    return int(
        Gtk.get_current_event_time()
        or keybindings.get_current_event_time()
        or _internal_data.current_event_time
    )


def _parse_notify_id(startup_notification_id: str) -> int:
    """Return timestamp or 0 from @startup_notification_id."""
    if "_TIME" in startup_notification_id:
        _ign, bstime = startup_notification_id.split("_TIME", 1)
        with contextlib.suppress(ValueError):
            return abs(int(bstime))

    return 0


@contextlib.contextmanager
def using_startup_notify_id(notify_id: str) -> ty.Any:
    """Pass in a DESKTOP_STARTUP_ID

    with using_startup_notify_id(...) as time:
        pass

    The yelt object is the parsed timestamp
    """
    if timestamp := _parse_notify_id(notify_id):
        Gdk.notify_startup_complete_with_id(notify_id)

    try:
        pretty.print_debug(__name__, "Using startup id", notify_id, timestamp)
        _internal_data.current_event_time = timestamp
        yield timestamp
    finally:
        _internal_data.current_event_time = Gdk.CURRENT_TIME



from __future__ import annotations

import gzip
import hashlib
import itertools
import os
import pickle
import threading
import time
import traceback
import typing as ty
import weakref
from collections import defaultdict
from pathlib import Path

from kupfer import config
from kupfer.core import pluginload, plugins
from kupfer.obj import Action, AnySource, Leaf, Source, TextSource
from kupfer.obj.sources import MultiSource, SourcesSource
from kupfer.support import conspickle, pretty, scheduler

if ty.TYPE_CHECKING:
    from kupfer.obj.base import ActionGenerator

__all__ = (
    "InternalError",
    "PeriodicRescanner",
    "SourceController",
    "SourceDataPickler",
    "SourcePickler",
    "get_source_controller",
)


class InternalError(Exception):
    pass


class PeriodicRescanner(pretty.OutputMixin):
    """Periodically rescan a @catalog of sources.

    Do first rescan after @startup seconds, then followup with rescans in @period.

    Each campaign of rescans is separarated by @campaign seconds.

    Source may define own min rescan interval, so can be rescanned less
    frequently.
    Last rescan is hold in source last_scan property, so when zeroed - force
    load on next rescan.
    """

    def __init__(
        self,
        period: int = 5,
        startup: int = 10,
        campaign: int = 60,
        min_rescan_interval: int = 900,
    ) -> None:
        self._startup = startup
        self._period = period
        self._campaign = campaign
        self._timer = scheduler.Timer()
        self._min_rescan_interval = min_rescan_interval
        self._catalog: ty.Iterable[Source] = []
        self._catalog_cur: ty.Iterator[Source] | None = None

    def set_catalog(self, catalog: ty.Iterable[Source]) -> None:
        self._catalog = catalog
        self._catalog_cur = iter(self._catalog)
        self.output_debug(f"Registering new campaign, in {self._startup} s")
        self._timer.set(self._startup, self._new_campaign)

    def _new_campaign(self) -> None:
        self.output_info(f"Starting new campaign, interval {self._period} s")
        self._catalog_cur = iter(self._catalog)
        self._timer.set(self._period, self._periodic_rescan_helper)

    def _periodic_rescan_helper(self) -> None:
        # Advance until we find a source that was not recently rescanned
        for source in self._catalog_cur or ():
            # skip dynamic sources
            if source.is_dynamic():
                continue

            # for old objects that may not have this attribute
            last_scan = getattr(source, "last_scan", 0)
            interval = max(
                getattr(source, "source_scan_interval", 0),
                self._min_rescan_interval,
            )

            next_scan = last_scan + interval - time.time()
            if next_scan <= 0:
                self.output_debug(f"scanning {source}")
                self._start_source_rescan(source)
                return

            self.output_debug(
                f"source {source} up to date, next scan in {int(next_scan)}s"
            )

        # No source to scan found
        self.output_info(f"Campaign finished, pausing {self._campaign} s")
        self._timer.set(self._campaign, self._new_campaign)

    def rescan_now(self, source: Source, force_update: bool = False) -> None:
        "Rescan @source immediately"
        if source.is_dynamic():
            return

        # on initial rescan force update also "cached" source when there is no
        # items in cache.
        force_update |= (
            isinstance(source.cached_items, (list, tuple))
            and len(source.cached_items) == 0
        )

        self._rescan_source(source, force_update=force_update, campaign=False)

    def _start_source_rescan(self, source: Source) -> None:
        thread = threading.Thread(
            target=self._rescan_source, args=(source, True, True)
        )
        thread.daemon = True
        thread.start()

    def _rescan_source(
        self, source: Source, force_update: bool = True, campaign: bool = False
    ) -> None:
        start = time.monotonic()
        try:
            leaves = source.get_leaves(force_update=force_update) or ()
            cnt = sum(1 for leaf in leaves)
            duration = time.monotonic() - start
            self.output_info(
                f"scan {source}: {cnt} leaves in {duration:0.5f} s"
            )
        except Exception as exc:
            self.output_error(f"Scanning {source}: {exc}")
            self.output_exc()

        if campaign:
            # try to process next source
            self._timer.set(self._period, self._periodic_rescan_helper)


class SourcePickler(pretty.OutputMixin):
    """Takes care of pickling and unpickling Kupfer Sources."""

    format_version = 5
    name_template = "k%s-v%d.pickle.gz"

    def __init__(self) -> None:
        # TODO: check
        self.open = lambda f, mode: gzip.open(f, mode, compresslevel=3)  # noqa:SIM115

    @classmethod
    def should_use_cache(cls) -> bool:
        return config.has_capability("CACHE")

    @classmethod
    def should_use_cache_for_source(cls, source: Source) -> bool:
        return cls.should_use_cache() and source.source_use_cache

    def rm_old_cachefiles(self) -> None:
        """Checks if there are old cachefiles from last version,
        and deletes those"""
        home = config.get_cache_home()
        assert home
        for dpath, _dirs, files in os.walk(home):
            # Look for files matching beginning and end of
            # name_template, with the previous file version
            chead, ctail = self.name_template.split("%s")
            ctail = ctail % ((self.format_version - 1),)
            obsolete_files = []
            for cfile in files:
                if cfile.startswith(chead) and cfile.endswith(ctail):
                    cfullpath = os.path.join(dpath, cfile)
                    obsolete_files.append(cfullpath)

        if obsolete_files:
            self.output_info(
                "Removing obsolete cache files:",
                *obsolete_files,
                sep="\n",
            )
            for fpath in obsolete_files:
                # be overly careful
                assert fpath.startswith(home)
                assert "kupfer" in fpath
                Path(fpath).unlink()

    def get_filename(self, source: Source) -> str:
        """Return cache filename for @source"""
        # make sure we take the source name into account
        # so that we get a "break" when locale changes
        source_id = f"{source!r}{source}{source.version}"
        hash_str = hashlib.md5(source_id.encode("utf-8")).hexdigest()
        filename = self.name_template % (hash_str, self.format_version)
        home = config.get_cache_home()
        assert home
        return os.path.join(home, filename)

    def unpickle_source(self, source: Source) -> ty.Any:
        if not self.should_use_cache_for_source(source):
            return None

        if cached := self._unpickle_source(self.get_filename(source)):
            # check consistency
            if source == cached:
                return cached

            self.output_debug("Cached version mismatches", source)

        return None

    def _unpickle_source(self, pickle_file: str) -> ty.Any:
        try:
            pfile = Path(pickle_file).read_bytes()
            source = pickle.loads(pfile)
            assert isinstance(source, Source), "Stored object not a Source"
            sname = os.path.basename
            self.output_debug("Loading", source, "from", sname(pickle_file))
            return source
        except OSError:
            return None
        except (pickle.PickleError, Exception) as exc:
            self.output_info(f"Error loading {pickle_file}: {exc}")

        return None

    def pickle_source(self, source: Source) -> bool:
        if self.should_use_cache_for_source(source):
            return self._pickle_source(self.get_filename(source), source)

        return False

    def _pickle_source(self, pickle_file: str, source: Source) -> bool:
        """When writing to a file, use pickle.dumps() and then write the file
        in one go. If the file is a gzip file, pickler's thousands  of small
        writes are very slow."""
        sname = os.path.basename
        self.output_debug("Storing", source, "as", sname(pickle_file))
        Path(pickle_file).write_bytes(
            pickle.dumps(source, pickle.HIGHEST_PROTOCOL)
        )
        return True


class SourceDataPickler(pretty.OutputMixin):
    """Takes care of pickling and unpickling Kupfer Sources' configuration
    or data.

    The SourceDataPickler requires a protocol of three methods:

    config_save_name()
      Return an ascii name to be used as a token/key for the configuration

    config_save()
      Return an object to be saved as configuration

    config_restore(obj)
      Receive the configuration object `obj' to load
    """

    format_version = 2
    name_template = "config-%s-v%d.pickle"

    def __init__(self) -> None:
        self.open = open

    @classmethod
    def get_filename(cls, source: Source) -> str:
        """Return filename for @source"""
        assert hasattr(source, "config_save_name")
        name = source.config_save_name()
        filename = cls.name_template % (name, cls.format_version)
        filepath = config.save_config_file(filename)
        assert filepath
        return filepath

    @classmethod
    def source_has_config(cls, source: Source) -> ty.Any:
        return getattr(source, "config_save_name", None)

    def load_source(self, source: Source) -> None:
        if data := self._load_data(self.get_filename(source)):
            assert hasattr(source, "config_restore")
            source.config_restore(data)

    def _load_data(self, pickle_file: str) -> ty.Any:
        try:
            pfile = Path(pickle_file).read_bytes()
            data = conspickle.BasicUnpickler.loads(pfile)
            sname = os.path.basename(pickle_file)
            self.output_debug("Loaded configuration from", sname)
            return data
            # self.output_debug(data)
        except OSError:
            return None
        except (pickle.PickleError, Exception) as exc:
            self.output_error(f"Loading {pickle_file}: {exc}")

        return None

    def save_source(self, source: Source) -> bool:
        return self._save_data(self.get_filename(source), source)

    def _save_data(self, pickle_file: str, source: Source) -> bool:
        sname = os.path.basename(pickle_file)
        assert hasattr(source, "config_save")
        obj = source.config_save()
        try:
            data = pickle.dumps(obj, pickle.HIGHEST_PROTOCOL)
        except pickle.PickleError:
            self.output_error("Unable to save configuration for", source)
            self.output_error("Saving configuration raised an exception:")
            traceback.print_exc()
            self.output_error("Please file a bug report")
            data = None

        if data:
            self.output_debug("Storing configuration for", source, "as", sname)
            ## Write to temporary and rename into place
            tmp_pickle_file = f"{pickle_file}.{os.getpid()}"
            Path(tmp_pickle_file).write_bytes(data)
            os.rename(tmp_pickle_file, pickle_file)

        return True


## define type for Source subclass to prevent problems with mixins
SourceSubclass = ty.TypeVar("SourceSubclass", bound=Source)


# pylint: disable=too-many-instance-attributes,too-many-public-methods
class SourceController(pretty.OutputMixin):
    """Control sources; loading, pickling, rescanning

    Call .add() to add sources.
    Call .initialize() before use commences.
    """

    _instance: SourceController | None = None

    @classmethod
    def instance(cls) -> SourceController:
        """Get instance of SourceController (singleton)"""
        if cls._instance is None:
            cls._instance = SourceController()

        assert cls._instance
        return cls._instance

    def __init__(self):
        self._sources: set[Source] = set()
        self.action_decorators: dict[ty.Type[Leaf], set[Action]] = defaultdict(
            set
        )
        self._rescanner = PeriodicRescanner(period=1)
        self._toplevel_sources: set[Source] = set()
        self._text_sources: set[TextSource] = set()
        self._content_decorators: dict[ty.Type[Leaf], set[ty.Type[Source]]] = (
            defaultdict(set)
        )
        self._action_generators: list[ActionGenerator] = []
        # map any object defined in plugin to plugin id
        self._plugin_object_map: weakref.WeakKeyDictionary[ty.Any, str] = (
            weakref.WeakKeyDictionary()
        )
        self._loaded_successfully = False
        self._did_finalize_sources = False
        # _pre_root cache root sources
        self._pre_root: list[Source] | None = None

    def add(
        self,
        plugin_id: str | None,
        srcs: ty.Iterable[Source],
        toplevel: bool = False,
        initialize: bool = False,
    ) -> None:
        self._invalidate_root()
        new_srcs = set(self._try_restore(srcs))
        new_srcs.update(srcs)

        self._sources.update(new_srcs)
        if toplevel:
            self._toplevel_sources.update(new_srcs)

        if initialize:
            self._initialize_sources(new_srcs)
            self._cache_sources(new_srcs)
            self._rescanner.set_catalog(self._sources)

        if plugin_id:
            self._register_plugin_objects(plugin_id, *new_srcs)

    def set_toplevel(self, src: Source, toplevel: bool) -> None:
        assert src in self._sources, (
            "Source is not tracked in SourceController"
        )
        self._invalidate_root()
        if toplevel:
            self._toplevel_sources.add(src)
        else:
            self._toplevel_sources.discard(src)

    def _register_plugin_objects(
        self, plugin_id: str, *objects: ty.Any
    ) -> None:
        """Register a plugin id mapping for @objects"""
        for obj in objects:
            self._plugin_object_map[obj] = plugin_id
            pretty.print_debug(__name__, "Add", obj)

    def _remove(self, src: Source) -> None:
        self._invalidate_root()
        self._toplevel_sources.discard(src)
        self._sources.discard(src)
        self._rescanner.set_catalog(self._sources)
        self._finalize_source(src)
        pretty.print_debug(__name__, "Remove", src)

    def get_plugin_id_for_object(self, obj: ty.Any) -> str | None:
        return self._plugin_object_map.get(obj)

    def remove_objects_for_plugin_id(self, plugin_id: str) -> bool:
        """Remove all objects for @plugin_id

        Return True if the catalog configuration changed
        """
        removed_source = False
        self.output_debug("Removing objects for plugin:", plugin_id)

        # sources
        for src in list(self._sources):
            if self.get_plugin_id_for_object(src) == plugin_id:
                self._remove(src)
                removed_source = True

        # all other objects
        def remove_matching_objects(collection, plugin_id):
            for obj in list(collection):
                if self.get_plugin_id_for_object(obj) == plugin_id:
                    collection.remove(obj)
                    pretty.print_debug(__name__, "Remove", obj)

        remove_matching_objects(self._text_sources, plugin_id)

        for typ_v in self._content_decorators.values():
            remove_matching_objects(typ_v, plugin_id)

        for a_typ_v in self.action_decorators.values():
            remove_matching_objects(a_typ_v, plugin_id)

        remove_matching_objects(self._action_generators, plugin_id)

        return removed_source

    def get_sources(self) -> set[Source]:
        return self._sources

    def add_text_sources(self, plugin_id, srcs):
        self._text_sources.update(srcs)
        self._register_plugin_objects(plugin_id, *srcs)

    def get_text_sources(self) -> set[TextSource]:
        return self._text_sources

    def add_content_decorators(
        self,
        plugin_id: str,
        decos: dict[ty.Type[Leaf], set[ty.Type[Source]]],
    ) -> None:
        for typ, val in decos.items():
            self._content_decorators[typ].update(val)
            self._register_plugin_objects(plugin_id, *val)

    def add_action_decorators(
        self, plugin_id: str, decos: dict[ty.Type[Leaf], list[Action]]
    ) -> None:
        for typ, val in decos.items():
            self.action_decorators[typ].update(val)
            self._register_plugin_objects(plugin_id, *val)

        for typ_v in self.action_decorators.values():
            self._disambiguate_actions(typ_v)

    def add_action_generator(
        self, plugin_id: str, agenerator: ActionGenerator
    ) -> None:
        self._action_generators.append(agenerator)
        self._register_plugin_objects(plugin_id, agenerator)

    def _disambiguate_actions(self, actions: ty.Iterable[Action]) -> None:
        """Rename actions by the same name (adding a suffix)"""
        names: dict[str, Action] = {}
        renames = set()
        for action in actions:
            name = str(action)
            if name in names:
                renames.add(names[name])
                renames.add(action)
            else:
                names[name] = action

        for action in renames:
            self.output_debug(f"Disambiguate Action {action}")
            plugin_name = plugins.get_plugin_name(type(action).__module__)
            plugin_suffix = f" ({plugin_name})"
            if not action.name.endswith(plugin_suffix):
                action.name += plugin_suffix

    def __contains__(self, src: Source) -> bool:
        return src in self._sources

    def __getitem__(self, src: Source) -> Source:
        for source in self._sources:
            if source == src:
                return source

        raise KeyError

    @property
    def root(self) -> Source | None:
        """Get the root source of catalog"""
        if len(self._sources) == 1:
            (root_catalog,) = self._sources  # pylint: disable=unbalanced-tuple-unpacking
        elif len(self._sources) > 1:
            root_catalog = MultiSource(self.firstlevel)
        else:
            root_catalog = None

        return root_catalog

    def _invalidate_root(self) -> None:
        "The source root needs to be recalculated"
        self._pre_root = None

    @property
    def firstlevel(self) -> list[Source]:
        if self._pre_root:
            return self._pre_root

        sourceindex = self._sources.copy()
        kupfer_sources = SourcesSource(self._sources)
        sourceindex.add(kupfer_sources)
        # Make sure firstlevel is ordered
        # So that it keeps the ordering.. SourcesSource first
        firstlevel: list[Source] = [SourcesSource(sourceindex)]
        firstlevel.extend(self._toplevel_sources)
        self._pre_root = firstlevel
        return firstlevel

    @classmethod
    def good_source_for_types(
        cls, source: AnySource, types: tuple[ty.Any, ...]
    ) -> bool:
        """return whether @s provides good leaves for @types"""
        if provides := list(source.provides()):
            return any(issubclass(t, types) for t in provides)

        return True

    def root_for_types(
        self,
        types: ty.Iterable[ty.Any],
        extra_sources: ty.Iterable[Source] | None = None,
    ) -> MultiSource:
        """Get root for a flat catalog of all catalogs providing at least
        Leaves of @types.

        types: Iterable of classes
        extra_sources: Sources to include

        Take all sources which:
            Provide a type T so that it is a subclass
            to one in the set of types we want
        """
        ttypes: tuple[ty.Any, ...] = tuple(types)
        firstlevel = set(extra_sources or [])
        # include the Catalog index since we want to include
        # the top of the catalogs (like $HOME)
        catalog_index = (SourcesSource(self._sources),)
        firstlevel.update(
            s
            for s in itertools.chain(self._sources, catalog_index)
            if self.good_source_for_types(s, ttypes)
        )

        return MultiSource(firstlevel)

    def get_canonical_source(self, src: AnySource) -> AnySource:
        """Return the canonical instance for @source"""
        # check if we already have source, then return that
        if not isinstance(src, Source):
            return src

        try:
            return self[src]
        except KeyError:
            src.initialize()
            return src

    def get_contents_for_leaf(
        self, leaf: Leaf, types: tuple[ty.Any, ...] | None = None
    ) -> ty.Iterator[AnySource]:
        """Iterator of content sources for @leaf, providing @types
        (or None for all)"""
        for typ, val in self._content_decorators.items():
            if not isinstance(leaf, typ):
                continue

            for content in val:
                with pluginload.exception_guard(
                    content, self._remove_source, content, is_decorator=True
                ):
                    dsrc = content.decorate_item(leaf)  # type: ignore

                if not dsrc:
                    continue

                if types and not self.good_source_for_types(dsrc, types):
                    continue

                yield self.get_canonical_source(dsrc)

    def get_actions_for_leaf(self, leaf: Leaf) -> ty.Iterator[Action]:
        for typ, val in self.action_decorators.items():
            if isinstance(leaf, typ):
                yield from val

        for agenerator in self._action_generators:
            yield from agenerator.get_actions_for_leaf(leaf)

    def decorate_object(self, obj: Leaf, action: Action | None = None) -> None:
        """If `obj` may have content (Source) and currently there is no assigned
        content - get sources for `obj` and optional `action` and add it into obj.
        Multiple sources are packed into SourcesSource.
        WARN: obj is updated in packet
        """
        if hasattr(obj, "has_content") and not obj.has_content():
            types = tuple(action.object_types()) if action else ()
            contents = tuple(self.get_contents_for_leaf(obj, types))

            if not contents:
                obj.add_content(None)
            elif len(contents) == 1:
                assert isinstance(contents[0], Source)
                obj.add_content(contents[0])
            else:
                assert isinstance(contents[0], Source)
                sources = ty.cast("ty.Collection[Source]", contents)
                obj.add_content(
                    SourcesSource(sources, name=str(obj), use_reprs=False)
                )

    def finalize(self) -> None:
        """Finalize all sources, equivalent to deactivating all sources"""
        for src in self._sources:
            src.finalize()

        self._did_finalize_sources = True

    def save_cache(self) -> None:
        """Save all caches (non-important data)"""
        if not self._did_finalize_sources:
            raise InternalError("Called save_cache without finalize!")

        if self._loaded_successfully:
            self._pickle_sources(self._sources)
        else:
            self.output_debug("Not writing cache on failed load")

    def save_data(self) -> None:
        """Save (important) user data/configuration"""
        if not self._loaded_successfully:
            self.output_info("Not writing configuration on failed load")
            return

        configsaver = SourceDataPickler()
        for source in self._sources:
            if configsaver.source_has_config(source):
                self._save_source(source, pickler=configsaver)

    @classmethod
    def _save_source(cls, source: Source, pickler: ty.Any = None) -> None:
        configsaver = pickler or SourceDataPickler()
        configsaver.save_source(source)

    def _finalize_source(self, source: Source) -> None:
        """Either save config, or save cache for @source"""
        source.finalize()
        if SourceDataPickler.source_has_config(source):
            self._save_source(source)
        elif not source.is_dynamic():
            self._pickle_source(source)

    def _pickle_sources(self, sources: ty.Iterable[Source]) -> None:
        sourcepickler = SourcePickler()
        sourcepickler.rm_old_cachefiles()
        for source in sources:
            if source.is_dynamic() or SourceDataPickler.source_has_config(
                source
            ):
                continue

            self._pickle_source(source, pickler=sourcepickler)

    @classmethod
    def _pickle_source(cls, source: Source, pickler: ty.Any = None) -> None:
        sourcepickler = pickler or SourcePickler()
        sourcepickler.pickle_source(source)

    def _try_restore(self, srcs: ty.Iterable[Source]) -> ty.Iterator[Source]:
        """Try to restore the source that is equivalent to the
        "dummy" instance @source, from cache, or from saved configuration.
        yield the instances that succeed.
        """
        sourcepickler = SourcePickler()
        configsaver = SourceDataPickler()
        for source in srcs:
            if configsaver.source_has_config(source):
                configsaver.load_source(source)
            else:
                source = sourcepickler.unpickle_source(source)  # noqa:PLW2901

            if source:
                yield source

    def _remove_source(
        self, source: Source, is_decorator: bool = False
    ) -> None:
        """Oust @source from catalog if any exception is raised"""
        if not is_decorator:
            self._sources.discard(source)
            self._toplevel_sources.discard(source)
            source_type = type(source)
        else:
            source_type = ty.cast("ty.Type[Source]", source)

        for cdv in self._content_decorators.values():
            cdv.discard(source_type)

    def initialize(self) -> None:
        """Initialize all sources and cache toplevel sources"""
        self._initialize_sources(self._sources)
        self._rescanner.set_catalog(self._sources)
        self._cache_sources(self._toplevel_sources)
        self._loaded_successfully = True

    def _initialize_sources(self, srcs: ty.Iterable[Source]) -> None:
        for src in srcs:
            with pluginload.exception_guard(src, self._remove_source, src):
                src.initialize()

    def _cache_sources(self, srcs: ty.Iterable[Source]) -> None:
        # Make sure that the toplevel sources are cached
        # either newly rescanned or the cache is fully loaded
        self.output_info("Initial sources load")
        for src in srcs:
            with pluginload.exception_guard(src, self._remove_source, src):
                self._rescanner.rescan_now(src, force_update=False)


get_source_controller = SourceController.instance



# Distributed under terms of the GPLv3 license.

"""
Test for datatools module.
"""

import unittest

from kupfer.support import datatools as d


class TestLruCache(unittest.TestCase):
    def test_insers(self):
        cache: d.LruCache[int, int] = d.LruCache(10)
        for i in range(5):
            cache[i] = i

        self.assertEqual(len(cache), 5)
        self.assertEqual(list(cache.keys()), [0, 1, 2, 3, 4])

        for i in range(10):
            cache[i] = i

        self.assertEqual(len(cache), 10)
        self.assertEqual(list(cache.keys()), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

        for i in range(20):
            cache[i] = i

        self.assertEqual(len(cache), 10)
        self.assertEqual(
            list(cache.keys()), [10, 11, 12, 13, 14, 15, 16, 17, 18, 19]
        )

        for i in range(10, 16):
            cache[i] = i

        self.assertEqual(len(cache), 10)
        self.assertEqual(
            list(cache.keys()), [16, 17, 18, 19, 10, 11, 12, 13, 14, 15]
        )

        for i in range(16, 20):
            self.assertEqual(cache[i], i)

        self.assertEqual(
            list(cache.keys()), [10, 11, 12, 13, 14, 15, 16, 17, 18, 19]
        )

    def test_get_or_insert(self):
        cache: d.LruCache[int, int] = d.LruCache(10)

        # pylint: disable=too-few-public-methods
        class Creator:
            def __init__(self):
                self.cntr = 0

            def call(self):
                self.cntr += 1
                return self.cntr

        creator = Creator()

        val = cache.get_or_insert(0, creator.call)
        self.assertEqual(val, 1)

        val = cache.get_or_insert(0, creator.call)
        self.assertEqual(val, 1)

        val = cache.get_or_insert(1, creator.call)
        self.assertEqual(val, 2)

        val = cache.get_or_insert(2, creator.call)
        self.assertEqual(val, 3)

        val = cache.get_or_insert(1, creator.call)
        self.assertEqual(val, 2)

        val = cache.get_or_insert(0, creator.call)
        self.assertEqual(val, 1)

        self.assertEqual(creator.cntr, 3)



"""
Application - related objects (leaves, sources)

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import os
import typing as ty
from contextlib import suppress

from gi.repository import Gio, GLib

from kupfer import icons, launch
from kupfer.obj.base import Action, Leaf, Source
from kupfer.obj.exceptions import InvalidDataError, OperationError
from kupfer.obj.helplib import FilesystemWatchMixin, PicklingHelperMixin
from kupfer.support import pretty
from kupfer.version import DESKTOP_ID

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from kupfer.core import commandexec

__all__ = (
    "AppLeaf",
    "AppLeafContentMixin",
    "ApplicationSource",
    "CloseAll",
    "Launch",
    "LaunchAgain",
)


class AppLeafContentMixin:
    """Mixin for Source that correspond one-to-one with a AppLeaf.

    This Mixin sees to that the Source is set as content for the application
    with id 'cls.appleaf_content_id', which may also be a sequence of ids.

    Source has to define the attribute appleaf_content_id and must
    inherit this mixin BEFORE the Source

    This Mixin defines:
    get_leaf_repr
    decorates_type,
    decorates_item
    """

    _cached_leaf_repr: AppLeaf | None

    @classmethod
    def get_leaf_repr(cls) -> AppLeaf | None:
        if not hasattr(cls, "_cached_leaf_repr"):
            cls._cached_leaf_repr = cls.__get_leaf_repr()

        return cls._cached_leaf_repr

    @classmethod
    def __get_appleaf_id_iter(cls) -> ty.Iterator[str]:
        assert hasattr(cls, "appleaf_content_id")
        if isinstance(cls.appleaf_content_id, str):
            yield cls.appleaf_content_id
        else:
            yield from cls.appleaf_content_id

    @classmethod
    def __get_leaf_repr(cls) -> AppLeaf | None:
        for appleaf_id in cls.__get_appleaf_id_iter():
            with suppress(InvalidDataError):
                return AppLeaf(app_id=appleaf_id)

        return None

    @classmethod
    def decorates_type(cls) -> ty.Type[Leaf]:
        return AppLeaf

    @classmethod
    def decorate_item(cls, leaf: Leaf) -> AppLeafContentMixin | None:
        if leaf == cls.get_leaf_repr():
            return cls()

        return None


class ApplicationSource(
    AppLeafContentMixin, Source, PicklingHelperMixin, FilesystemWatchMixin
):
    """Abstract, helper class that include parent object to create source that
    is bound to application leaf with file/folders monitoring."""


class AppLeaf(Leaf):
    def __init__(
        self,
        item: ty.Any = None,
        init_path: str | None = None,
        app_id: str | None = None,
        require_x: bool = True,
        app_actions: list[tuple[str, str]] | None = None,
        *,
        load_extra_aliases: bool = False,
    ) -> None:
        """Try constructing an Application for GAppInfo @item,
        for file @path or for package name @app_id.

        @require_x: require executable file
        @app_actions: additional application actions defined in
           desktop file as list of (action name, action display name)

        Represented object is Gio.DesktopAppInfo.
        """
        self._init_path = init_path
        self._init_item_id = app_id and app_id + ".desktop"
        # finish will raise InvalidDataError on invalid item
        self._finish(require_x, item)
        self.app_actions = app_actions
        super().__init__(self.object, self.object.get_name())
        self._add_aliases(load_extra_aliases)

    def _add_aliases(self, load_extra_aliases: bool) -> None:
        # find suitable alias
        # use package name: non-extension part of ID
        package_name = GLib.filename_display_basename(self.get_id())
        if package_name and package_name not in str(self).lower():
            self.kupfer_add_alias(package_name)

        # add executable as alias
        if cmdl := self.object.get_executable():
            self.kupfer_add_alias(cmdl)

        # do not continue when application setting "load extra aliases" if off
        if not load_extra_aliases:
            return

        # add non-localized name
        if (en_name := self.object.get_string("Name")) != self.name:
            self.kupfer_add_alias(en_name)

        # add generic name
        if (
            name := (
                self.object.get_string("GenericName")
                or self.object.get_string("GenericName")
            )
        ) and name != self.name:
            self.kupfer_add_alias(name)

        # add keywords
        if keywords := self.object.get_string_list("Keywords"):
            for keyword in keywords:
                self.kupfer_add_alias(keyword)

        # add Comment
        if comment := (
            self.object.get_locale_string("Comment")
            or self.object.get_string("Comment")
        ):
            self.kupfer_add_alias(comment)

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, type(self)) and self.get_id() == other.get_id()
        )

    def __getstate__(self) -> dict[str, ty.Any]:
        self._init_item_id = self.object and self.object.get_id()
        state = dict(vars(self))
        state["object"] = None
        return state

    def __setstate__(self, state: dict[str, ty.Any]) -> None:
        vars(self).update(state)
        self._finish()

    def _finish(
        self,
        require_x: bool = False,
        item: Gio.DesktopAppInfo | None = None,
    ) -> None:
        """Try to set self.object from init's parameters"""
        if not item:
            # Construct an AppInfo item from either path or item_id
            try:
                if self._init_path and (
                    not require_x or os.access(self._init_path, os.X_OK)
                ):
                    # serilizable if created from a "loose file"
                    self.serializable = 1
                    item = Gio.DesktopAppInfo.new_from_filename(
                        self._init_path
                    )
                elif self._init_item_id:
                    item = Gio.DesktopAppInfo.new(self._init_item_id)

            except TypeError as exc:
                pretty.print_debug(
                    __name__,
                    "Application not found:",
                    self._init_item_id,
                    self._init_path,
                )
                raise InvalidDataError from exc

        self.object = item
        if not self.object:
            raise InvalidDataError

    def repr_key(self) -> ty.Any:
        return self.get_id()

    def launch(
        self,
        files: ty.Iterable[Gio.File] = (),
        paths: ty.Iterable[str] = (),
        activate: bool = False,
        ctx: commandexec.ExecutionToken | None = None,
        work_dir: str | None = None,
        uris: ty.Iterable[str] = (),
    ) -> bool:
        """Launch the represented applications.

        `files` is a sequence of GFiles (Gio.File), `paths` is a sequence of
        paths (str), `uris` is list of uri. All three lists are combined
        together. In most cases only one list should be passed to this method.

        When `activate` is True - activate running instance instead of start
        new.

        `work_dir` can overwrite work directory of application.
        """
        try:
            return launch.launch_application(
                self.object,
                files=files,
                paths=paths,
                uris=uris,
                activate=activate,
                desktop_file=self._init_path,
                screen=ctx and ctx.environment.get_screen(),
                work_dir=work_dir,
            )
        except launch.SpawnError as exc:
            raise OperationError(exc) from exc

    def get_id(self) -> str:
        """Return the unique ID for this app.

        This is the GIO id "gedit.desktop" minus the .desktop part for
        system-installed applications.
        """
        return launch.application_id(self.object, self._init_path)

    def get_actions(self) -> ty.Iterable[Action]:
        id_ = self.get_id()
        if id_ == DESKTOP_ID:
            return

        if launch.application_is_running(id_):
            yield Launch(_("Go To"), is_running=True)
            yield CloseAll()
        else:
            yield Launch()

        yield LaunchAgain()

    def get_description(self) -> str | None:
        # Use Application's description, else use executable
        # for "file-based" applications we show the path
        app_desc = self.object.get_description()
        ret: str = app_desc or self.object.get_executable()
        if self._init_path:
            app_path = launch.get_display_path_for_bytestring(self._init_path)
            return f"({app_path}) {ret}"

        return ret

    def get_gicon(self) -> icons.GIcon | None:
        return self.object.get_icon()

    def get_icon_name(self) -> str:
        return "exec"


class Launch(Action):
    """Launches an application (AppLeaf)"""

    action_accelerator: str | None = "o"
    rank_adjust = 5

    def __init__(
        self,
        name: str | None = _("Launch"),
        is_running: bool = False,
        open_new: bool = False,
    ) -> None:
        """
        If @is_running, style as if the app is running (Show application)
        If @open_new, always start a new instance.
        """
        Action.__init__(self, name)
        self.is_running = is_running
        self.open_new = open_new

    def wants_context(self) -> bool:
        return True

    def activate(
        self,
        leaf: Leaf,
        iobj: Leaf | None = None,
        ctx: commandexec.ExecutionToken | None = None,
    ) -> Leaf | None:
        assert isinstance(leaf, AppLeaf)
        leaf.launch(activate=not self.open_new, ctx=ctx)
        return None

    def get_description(self) -> str:
        if self.is_running:
            return _("Show application window")

        return _("Launch application")

    def get_icon_name(self) -> str:
        if self.is_running:
            return "go-jump"

        return "kupfer-launch"


class LaunchAgain(Launch):
    action_accelerator: str | None = None
    rank_adjust = 0

    def __init__(self, name: str | None = _("Launch Again")):
        Launch.__init__(self, name, open_new=True)

    def item_types(self) -> ty.Iterator[ty.Type[Leaf]]:
        yield AppLeaf

    def valid_for_item(self, leaf: Leaf) -> bool:
        assert isinstance(leaf, AppLeaf)
        return launch.application_is_running(leaf.get_id())

    def get_description(self) -> str:
        return _("Launch another instance of this application")


class CloseAll(Action):
    """Attempt to close all application windows"""

    rank_adjust = -10

    def __init__(self, name: str | None = _("Close")) -> None:
        Action.__init__(self, name)

    def activate(
        self,
        leaf: Leaf,
        iobj: Leaf | None = None,
        ctx: commandexec.ExecutionToken | None = None,
    ) -> Leaf | None:
        assert isinstance(leaf, AppLeaf)
        launch.application_close_all(leaf.get_id())
        return None

    def item_types(self) -> ty.Iterator[ty.Type[Leaf]]:
        yield AppLeaf

    def valid_for_item(self, leaf: Leaf) -> bool:
        assert isinstance(leaf, AppLeaf)
        return launch.application_is_running(leaf.get_id())

    def get_description(self) -> str:
        return _("Attempt to close all application windows")

    def get_icon_name(self) -> str:
        return "window-close"



"""
This is a DuckDuckGo search plugin based on the Wikipedia search plugin
"""

__kupfer_name__ = _("DuckDuckGo Search")
__kupfer_sources__ = ()
__kupfer_actions__ = ("DuckDuckGoSearch",)
__description__ = _("Search the web securely with DuckDuckGo")
__version__ = "1.0"
__author__ = "Isaac Aggrey <isaac.aggrey@gmail.com>"

import typing as ty
import urllib.parse

from kupfer import launch
from kupfer.obj import Action, TextLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class DuckDuckGoSearch(Action):
    def __init__(self):
        Action.__init__(self, _("DuckDuckGo Search"))

    def activate(self, leaf, iobj=None, ctx=None):
        search_url = "https://duckduckgo.com/"
        query_url = (
            search_url + "?" + urllib.parse.urlencode({"q": leaf.object})
        )
        launch.show_url(query_url)

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Search the web securely with DuckDuckGo")

    def get_icon_name(self):
        return "edit-find"



"""
Special objects.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import typing as ty

from kupfer.obj.objects import RunnableLeaf
from kupfer.ui import preferences

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from kupfer.core import commandexec

__version__ = "2023-09-09"
__all__ = (
    "CommandNotAvailableLeaf",
    "InvalidCredentialsLeaf",
    "NonfunctionalLeaf",
    "PleaseConfigureLeaf",
)


class PleaseConfigureLeaf(RunnableLeaf):
    """Show information and allow to open preferences for given plugin"""

    message = _("Please Configure Plugin")
    description = _("Plugin %s is not configured")

    def __init__(self, plugin_id: str, plugin_name: str) -> None:
        plugin_id = plugin_id.split(".")[-1]
        RunnableLeaf.__init__(self, plugin_id, self.message)
        self.plugin_name = plugin_name

    def wants_context(self) -> bool:
        return True

    def run(self, ctx: commandexec.ExecutionToken | None = None) -> None:
        assert ctx
        preferences.show_plugin_info(self.object, ctx.environment)

    def get_icon_name(self) -> str:
        return "preferences-desktop"

    def get_description(self) -> str:
        return self.description % self.plugin_name


class InvalidCredentialsLeaf(PleaseConfigureLeaf):
    description = _("Invalid user credentials for %s")


class NonfunctionalLeaf(PleaseConfigureLeaf):
    """Leaf with custom error message that open plugin preferences dialog."""

    def __init__(
        self, plugin_id: str, plugin_name: str, description: str
    ) -> None:
        super().__init__(plugin_id, plugin_name)
        self.description = description

    def get_description(self) -> str:
        return self.description


class CommandNotAvailableLeaf(PleaseConfigureLeaf):
    """Leaf with message "command ... not available" that open plugin
    preferences dialog."""

    description = _("Command '%s' not available")

    def __init__(self, plugin_id: str, plugin_name: str, command: str) -> None:
        super().__init__(plugin_id, plugin_name)
        self.command = command

    def get_description(self) -> str:
        return self.description % self.command



"""
virtualbox_ose_support.py

Control VirtualBox via command-line interface.
Support both Sun VirtualBox and VirtualBox OpenSource Edition.
"""

__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"
__version__ = "2018-10-21"

import os
from pathlib import Path
from xml.dom import minidom

from kupfer import launch
from kupfer.plugin.virtualbox import constants as vbox_const
from kupfer.support import pretty

_VBOX_CONFIG_DIRS = (
    Path("~/.config/VirtualBox/").expanduser(),
    Path("~/.VirtualBox/").expanduser(),
)
_VBOX_CONFIG_FILE = "VirtualBox.xml"

MONITORED_DIRS = _VBOX_CONFIG_DIRS
IS_DYNAMIC = False
ICON = "virtualbox-ose"
APP_ID = "virtualbox-ose"

# parameters for VBoxManage
_ACTIONS = {
    vbox_const.VM_POWEROFF: "poweroff",
    vbox_const.VM_ACPI_POWEROFF: "acpipowerbutton",
    vbox_const.VM_PAUSE: "pause",
    vbox_const.VM_REBOOT: "reset",
    vbox_const.VM_RESUME: "resume",
    vbox_const.VM_SAVE: "savestate",
}


def get_machine_state(vm_uuid):
    """check vms state (on/off/paused)"""
    state = vbox_const.VM_STATE_POWEROFF
    try:
        str_state = "poweroff"
        with os.popen(
            f"VBoxManage showvminfo {vm_uuid} --machinereadable"
        ) as pinfo:
            for line in pinfo:
                if line.startswith('VMState="'):
                    str_state = line.strip()[9:-1]
                    break

        if str_state == "paused":
            state = vbox_const.VM_STATE_PAUSED
        elif str_state == "running":
            state = vbox_const.VM_STATE_POWERON
        elif str_state == "saved":
            state = vbox_const.VM_STATE_SAVED

    except OSError as err:
        pretty.print_error(
            __name__, "get_machine_state", vm_uuid, "error", err
        )
        state = vbox_const.VM_STATE_POWEROFF

    return state


def vm_action(action, vm_uuid):
    """change state of the virtual machine. Call VBoxManage.
    @param action - one of the const VM_*
    @param vm_uuid - virtual machine uuid
    """
    if action == vbox_const.VM_START_NORMAL:
        launch.spawn_async(["VBoxManage", "startvm", vm_uuid, "--type", "gui"])
    elif action == vbox_const.VM_START_HEADLESS:
        launch.spawn_async(
            ["VBoxManage", "startvm", vm_uuid, "--type", "headless"]
        )
    else:
        command = _ACTIONS[action]
        launch.spawn_async(["VBoxManage", "controlvm", vm_uuid, command])


def _get_virtual_machines(config_file):
    """load (virtual machine uuid, path to vm config) from virtualbox
    configuration.
    @param config_file - path to VirtualBox.xml file
    """
    try:
        dtree = minidom.parse(config_file)
        machine_registry = dtree.getElementsByTagName("MachineRegistry")[0]
        for machine in machine_registry.getElementsByTagName("MachineEntry"):
            yield (
                machine.getAttribute("uuid")[1:-1],
                machine.getAttribute("src"),
            )

    except Exception as err:
        pretty.print_error(
            __name__, "_get_virtual_machines", config_file, "error", err
        )


def _get_machine_info(vm_uuid, config_file):
    """load information about virtual machines from its configuration file.
    @param vm_uuid - uuid virtual machine
    @param config_file - path to vm configuration file
    """
    if not os.path.isfile(config_file):
        return None, None

    try:
        dtree = minidom.parse(config_file)
        machine_registry = dtree.getElementsByTagName("Machine")[0]
        os_type = machine_registry.getAttribute("OSType")
        name = machine_registry.getAttribute("name")
        description = None
        for machine_registry_child in machine_registry.childNodes:
            if machine_registry_child.nodeName == "Description":
                if machine_registry_child.hasChildNodes() and (
                    fc := machine_registry_child.firstChild):
                        description = fc.nodeValue

                break

        return (name, description or os_type)

    except Exception as err:
        pretty.print_error(
            __name__, "_get_machine_info", vm_uuid, "error" + config_file, err
        )
    return None, None


def get_machines():
    for dirname in _VBOX_CONFIG_DIRS:
        vbox_conf = Path(dirname, _VBOX_CONFIG_FILE)
        if not vbox_conf.is_file():
            continue

        for vm_uuid, config in _get_virtual_machines(vbox_conf):
            if not os.path.isabs(config):
                config = os.path.join(dirname, config)  # noqa: PLW2901

            name, description = _get_machine_info(vm_uuid, config)
            if name:
                yield (vm_uuid, name, description)


def unload():
    pass



from __future__ import annotations

import typing as ty

from gi.repository import GLib, GObject

from kupfer.support import pretty, weaklib

__all__ = ("Scheduler", "Timer", "get_scheduler")


def get_scheduler() -> Scheduler:
    """Get the shared instance"""
    return Scheduler.instance()


class Scheduler(GObject.GObject, pretty.OutputMixin):  # type:ignore
    __gtype_name__ = "Scheduler"

    _instance: Scheduler | None = None

    @classmethod
    def instance(cls) -> Scheduler:
        if cls._instance is None:
            cls._instance = Scheduler()

        return cls._instance

    def __init__(self) -> None:
        super().__init__()
        self._finished = False

    def load(self) -> None:
        self.output_debug("Loading")
        self.emit("load")
        self.emit("loaded")
        self.output_debug("Loaded")

    def display(self) -> None:
        self.output_debug("Display")
        self.emit("display")
        GLib.idle_add(self._after_display)

    def _after_display(self) -> None:
        self.output_debug("After Display")
        self.emit("after-display")

    @property
    def finished(self) -> bool:
        return self._finished

    def finish(self) -> None:
        self._finished = True
        self.emit("finish")


GObject.signal_new(
    "load", Scheduler, GObject.SignalFlags.RUN_LAST, GObject.TYPE_BOOLEAN, ()
)
GObject.signal_new(
    "loaded", Scheduler, GObject.SignalFlags.RUN_LAST, GObject.TYPE_BOOLEAN, ()
)
GObject.signal_new(
    "display",
    Scheduler,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (),
)
GObject.signal_new(
    "after-display",
    Scheduler,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (),
)
GObject.signal_new(
    "finish", Scheduler, GObject.SignalFlags.RUN_LAST, GObject.TYPE_BOOLEAN, ()
)


TimerCallback = ty.Callable[..., None]


class Timer:
    def __init__(self, call_at_finish: bool = False) -> None:
        self._current_timer = -1
        self._call_at_finish = call_at_finish
        self._current_callback: ty.Callable[[], None] | None = None
        weaklib.gobject_connect_weakly(
            get_scheduler(), "finish", self._on_finish
        )

    def set(
        self,
        timeout_seconds: float,
        callback: TimerCallback,
        *arguments: ty.Any,
    ) -> None:
        """Setup timer to call @timeout_seconds in the future.
        If the timer was previously set, it is postponed
        """
        self.invalidate()
        self._current_callback = lambda: callback(*arguments)
        self._current_timer = GLib.timeout_add_seconds(
            timeout_seconds, self._call
        )

    def set_ms(
        self,
        timeout_milliseconds: int,
        callback: TimerCallback,
        *arguments: ty.Any,
    ) -> None:
        """Setup timer to call @timeout_milliseconds in the future.
        If the timer was previously set, it is postponed
        """
        self.invalidate()
        self._current_callback = lambda: callback(*arguments)
        self._current_timer = GLib.timeout_add(
            timeout_milliseconds, self._call
        )

    def set_idle(self, callback: TimerCallback, *arguments: ty.Any) -> None:
        self.invalidate()
        self._current_callback = lambda: callback(*arguments)
        self._current_timer = GLib.idle_add(self._call)

    def _call(self, timer: ty.Any = None) -> None:
        assert self._current_callback

        self._current_timer = -1
        self._current_callback()

    def invalidate(self) -> None:
        if self._current_timer > 0:
            GLib.source_remove(self._current_timer)

        self._current_timer = -1

    def is_valid(self) -> bool:
        """If Timer is currently set"""
        return self._current_timer > 0

    def _on_finish(self, _scheduler: ty.Any) -> None:
        if self._call_at_finish and self.is_valid():
            self._call()
        else:
            self.invalidate()



from __future__ import annotations

import ast
import configparser
import copy
import locale
import os
import typing as ty

from gi.repository import GLib, GObject, Pango

from kupfer import config
from kupfer.support import pretty, scheduler

__all__ = (
    "SettingsController",
    "get_configured_terminal",
    "get_settings_controller",
    "is_known_terminal_executable",
)

AltValidator = ty.Callable[[dict[str, ty.Any]], bool]
Config = dict[str, dict[str, ty.Any]]


@ty.runtime_checkable
class ExtendedSetting(ty.Protocol):
    """Protocol that define non-simple configuration option"""

    def load(
        self, plugin_id: str, key: str, config_value: str | float | None
    ) -> None:
        """load value for @plugin_id and @key, @config_value is value
        stored in regular Kupfer config for plugin/key"""

    def save(self, plugin_id: str, key: str) -> str:
        """Save value for @plugin_id and @key.
        @Return value that should be stored in Kupfer config for
        plugin/key (string)"""


PlugConfigValue = ty.Union[
    str, bool, int, float, list[ty.Any], ExtendedSetting
]


# pylint: disable=too-few-public-methods
@ty.runtime_checkable
class ValueConverter(ty.Protocol):
    """Protocol that represent callable used to convert value stored
    in config file (as str) into required value (int, float, etc).
    """

    def __call__(self, value: str, default: ty.Any) -> PlugConfigValue: ...


# PlugConfigValueType = ty.Union[type[PlugConfigValue], ValueConverter]
PlugConfigValueType = ty.Union[ty.Any, ValueConverter]


def _strbool(value: ty.Any, default: bool = False) -> bool:
    """Coerce bool from string value or bool"""
    if isinstance(value, bool):
        return value

    value = str(value).lower()
    if value in ("no", "false"):
        return False

    if value in ("yes", "true"):
        return True

    return default


def _strint(value: ty.Any, default: int = 0) -> int:
    """Coerce bool from string value or bool"""
    try:
        return int(value)
    except ValueError:
        return default


def _strlist(value: str, default: list[ty.Any] | None = None) -> list[ty.Any]:
    """Parse string into list using ast literal_eval.

    literal_eval handle only 'safe' data, so should work fine.
    """
    try:
        val = ast.literal_eval(value)
        if isinstance(val, list):
            return val

        raise ValueError(f"invalid type: {val!r}")
    except (TypeError, SyntaxError, MemoryError, RecursionError) as err:
        raise ValueError(f"evaluate {value!r} error") from err


def _override_encoding(name: str) -> str | None:
    """Return a new encoding name if we want to override it, else return None.

    This is used to “upgrade” ascii to UTF-8 since the latter is a superset.
    """
    if name.lower() in ("ascii", "ANSI_X3.4-1968".lower()):
        return "UTF-8"

    return None


def _fill_parser_read(
    parser: configparser.RawConfigParser,
    defaults: Config,
) -> None:
    """Add values from `defaults` to `parser`."""
    for secname, section in defaults.items():
        if not parser.has_section(secname):
            parser.add_section(secname)

        for key, default in section.items():
            if isinstance(default, (tuple, list)):
                default = SettingsController.sep.join(default)  # noqa:PLW2901

            elif isinstance(default, int):
                default = str(default)  # noqa:PLW2901

            parser.set(secname, key, default)


# pylint: disable=too-many-return-statements
def _parse_value(defval: ty.Any, value: str) -> ty.Any:  # noqa:PLR0911
    if isinstance(defval, tuple):
        if not value:
            return ()

        return tuple(
            filter(None, map(str.strip, value.split(SettingsController.sep)))
        )

    if isinstance(defval, list):
        if not value:
            return []

        return list(
            filter(None, map(str.strip, value.split(SettingsController.sep)))
        )

    if isinstance(defval, bool):
        return _strbool(value)

    if isinstance(defval, int):
        return type(defval)(value)

    return value


def _fill_confmap_from_parser(
    parser: configparser.RawConfigParser,
    confmap: Config,
    defaults: Config,
) -> None:
    """Put values from `parser` to `confmap` using `defaults` as schema."""
    for secname in parser.sections():
        if secname not in confmap:
            confmap[secname] = {}

        for key in parser.options(secname):
            value = parser.get(secname, key)
            if ((sec := defaults.get(secname)) is not None) and (
                (defval := sec.get(key)) is not None
            ):
                value = _parse_value(defval, value)

            confmap[secname][key] = value


def _confmap_difference(conf: Config, defaults: Config) -> Config:
    """Extract the non-default keys to write out"""
    difference = {}
    for secname, section in conf.items():
        if secname not in defaults:
            difference[secname] = section.copy()
            continue

        difference[secname] = {}
        for key, config_val in section.items():
            if (
                secname in defaults
                and key in defaults[secname]
                and defaults[secname][key] == config_val
            ):
                continue

            difference[secname][key] = config_val

        if not difference[secname]:
            difference.pop(secname)

    return difference


def _fill_parser_from_config(
    parser: configparser.RawConfigParser, defaults: Config
) -> None:
    """Put content of `defaults` into `parser`."""
    for secname in sorted(defaults):
        section = defaults[secname]
        if not parser.has_section(secname):
            parser.add_section(secname)

        for key in sorted(section):
            value = section[key]
            if isinstance(value, (tuple, list)):
                value = SettingsController.sep.join(value)

            elif isinstance(value, int):
                value = str(value)

            parser.set(secname, key, value)


# pylint: disable=too-many-public-methods
class SettingsController(GObject.GObject, pretty.OutputMixin):  # type: ignore
    __gtype_name__ = "SettingsController"
    config_filename = "kupfer.cfg"
    defaults_filename = "defaults.cfg"
    sep = ";"
    default_directories = (
        "~/",
        "~/Desktop",
    )
    # Minimal "defaults" to define all fields
    # Read defaults defined in a defaults.cfg file
    _defaults: ty.Final[dict[str, ty.Any]] = {
        "Kupfer": {
            "keybinding": "",
            "magickeybinding": "",
            "showstatusicon": True,
            "showstatusicon_ai": False,
            "usecommandkeys": True,
            "score_without_key": True,
        },
        "Appearance": {
            "icon_large_size": 128,
            "icon_small_size": 24,
            "list_height": 200,
        },
        "Directories": {
            "direct": default_directories,
            "catalog": (),
        },
        "DeepDirectories": {
            "direct": (),
            "catalog": (),
            "depth": 1,
        },
        "Keybindings": {},
        "Tools": {},
    }

    _inst: SettingsController | None = None

    @classmethod
    def instance(cls) -> SettingsController:
        """SettingsController is singleton; instance return one, global
        instance."""
        if cls._inst is None:
            cls._inst = SettingsController()

        assert cls._inst
        return cls._inst

    def __init__(self) -> None:
        GObject.GObject.__init__(self)
        self._defaults_path: str | None = None
        self.encoding = _override_encoding(locale.getpreferredencoding())
        self.output_debug("Using", self.encoding)
        self._config = self._read_config()
        self._save_timer = scheduler.Timer(True)
        self._alternatives: dict[str, ty.Any] = {}
        self._alternative_validators: dict[str, AltValidator | None] = {}

    def _update_config_save_timer(self) -> None:
        self._save_timer.set(60, self._save_config)

    def _read_config(self, read_config: bool = True) -> Config:
        """Read cascading config files: default -> then config
        (in all XDG_CONFIG_DIRS)."""
        parser = configparser.RawConfigParser()

        # Set up defaults
        confmap = copy.deepcopy(self._defaults)
        _fill_parser_read(parser, confmap)

        # Read all config files
        config_files: list[str] = []
        try:
            defaults_path = config.get_data_file(self.defaults_filename)
        except config.ResourceLookupError:
            self.output_error(
                f"Error: no default config file {self.defaults_filename} "
                "found!"
            )
        else:
            self._defaults_path = defaults_path
            config_files.append(defaults_path)

        if read_config and (
            config_path := config.get_config_file(self.config_filename)
        ):
            config_files.append(config_path)

        for config_file in config_files:
            try:
                parser.read(config_file, encoding=self.encoding)
            except OSError as exc:
                self.output_error(
                    f"Error reading configuration file {config_file}: {exc}"
                )
            except UnicodeDecodeError as exc:
                self.output_error(
                    f"Error reading configuration file {config_file}: {exc}"
                )

        # Read parsed file into the dictionary again
        _fill_confmap_from_parser(parser, confmap, self._defaults)
        return confmap

    def _save_config(self, _scheduler: ty.Any = None) -> None:
        self.output_debug("Saving config")
        config_path = config.save_config_file(self.config_filename)
        if not config_path:
            self.output_info("Unable to save settings, can't find config dir")
            return

        # read in just the default values
        default_confmap = self._read_config(read_config=False)
        parser = configparser.RawConfigParser()
        confmap = _confmap_difference(self._config, default_confmap)
        _fill_parser_from_config(parser, confmap)
        ## Write to tmp then rename over for it to be atomic
        temp_config_path = f"{config_path}.{os.getpid()}"
        with open(temp_config_path, "w", encoding="UTF_8") as out:
            parser.write(out)

        os.rename(temp_config_path, config_path)

    def get_config(self, section: str, key: str) -> ty.Any:
        """General interface, but section must exist"""
        if section in self._defaults:
            key = key.lower()
            return self._config[section].get(key)

        raise KeyError(f"Invalid settings section: {section}")

    def set_config(self, section: str, key: str, value: ty.Any) -> bool:
        """General interface, but section must exist"""
        self.output_debug("Set", section, key, "to", value)
        key = key.lower()
        oldvalue = self._config[section].get(key)
        if section in self._defaults:
            if oldvalue is None:
                self._config[section][key] = str(value)
            else:
                self._config[section][key] = type(oldvalue)(value)

            self._emit_value_changed(section, key, value)
            self._update_config_save_timer()
            return True

        raise KeyError(f"Invalid settings section: {section}")

    def _emit_value_changed(
        self, section: str, key: str, value: ty.Any
    ) -> None:
        signal = f"value-changed::{section.lower()}.{key.lower()}"
        self.emit(signal, section, key, value)

    def _get_raw_config(
        self, section: str, key: str
    ) -> str | bool | int | float | None:
        """General interface, but section must exist"""
        key = key.lower()
        return self._config[section].get(key)

    def _set_raw_config(
        self, section: str, key: str, value: str | bool | float | None
    ) -> bool:
        """General interface, but will create section"""
        self.output_debug("Set", section, key, "to", value)
        if section not in self._config:
            self._config[section] = {}

        key = key.lower()
        self._config[section][key] = value
        self._update_config_save_timer()
        return False

    def _get_from_defaults(
        self, section: str, option: str
    ) -> str | bool | int | float | None:
        """Load values from default configuration file."""
        if self._defaults_path is None:
            self.output_error("Defaults not found")
            return None

        parser = configparser.RawConfigParser()
        parser.read(self._defaults_path)
        return parser.get(section, option.lower())

    def _get_from_defaults_section(
        self, section: str
    ) -> list[tuple[str, str]] | None:
        """Load values from default configuration file, return all section
        items as (key, value)."""
        if self._defaults_path is None:
            self.output_error("Defaults not found")
            return None

        parser = configparser.RawConfigParser()
        parser.read(self._defaults_path)
        return parser.items(section)

    def get_config_int(self, section: str, key: str) -> int:
        """section must exist"""
        if section not in self._defaults:
            raise KeyError(f"Invalid settings section: {section}")

        key = key.lower()
        value = self._config[section].get(key)
        return _strint(value)

    def get_plugin_enabled(self, plugin_id: str) -> bool:
        """Convenience: if @plugin_id is enabled"""
        return self.get_plugin_config_bool(plugin_id, "kupfer_enabled", False)

    def set_plugin_enabled(self, plugin_id: str, enabled: bool) -> bool:
        """Convenience: set if @plugin_id is enabled"""
        ret = self.set_plugin_config(
            plugin_id, "kupfer_enabled", enabled, value_type=_strbool
        )
        self.emit("plugin-enabled-changed", plugin_id, enabled)
        return ret

    def get_plugin_is_hidden(self, plugin_id: str) -> bool:
        """Convenience: if @plugin_id is hidden"""
        return self.get_plugin_config_bool(plugin_id, "kupfer_hidden", False)

    @classmethod
    def _source_config_repr(cls, obj: ty.Any) -> str:
        name = type(obj).__name__
        return "".join((c if c.isalnum() else "_") for c in name)

    def get_source_is_toplevel(self, plugin_id: str, src: ty.Any) -> bool:
        key = "kupfer_toplevel_" + self._source_config_repr(src)
        default = not getattr(src, "source_prefer_sublevel", False)
        return self.get_plugin_config_bool(plugin_id, key, default)

    def set_source_is_toplevel(
        self, plugin_id: str, src: ty.Any, value: bool
    ) -> bool:
        key = "kupfer_toplevel_" + self._source_config_repr(src)
        self.emit("plugin-toplevel-changed", plugin_id, value)
        return self.set_plugin_config(
            plugin_id, key, value, value_type=_strbool
        )

    def get_keybinding(self) -> str:
        """Convenience: Kupfer keybinding as string"""
        return str(self.get_config("Kupfer", "keybinding"))

    def set_keybinding(self, keystr: str) -> bool:
        """Convenience: Set Kupfer keybinding as string"""
        return self.set_config("Kupfer", "keybinding", keystr)

    def get_magic_keybinding(self) -> str:
        """Convenience: Kupfer alternate keybinding as string"""
        return str(self.get_config("Kupfer", "magickeybinding"))

    def set_magic_keybinding(self, keystr: str) -> bool:
        """Convenience: Set alternate keybinding as string"""
        return self.set_config("Kupfer", "magickeybinding", keystr)

    def get_global_keybinding(self, key: str) -> str:
        if key == "keybinding":
            return self.get_keybinding()

        if key == "magickeybinding":
            return self.get_magic_keybinding()

        raise ValueError("invalid key {key}")

    def set_global_keybinding(self, key: str, val: str) -> bool:
        if key == "keybinding":
            return self.set_keybinding(val)

        if key == "magickeybinding":
            return self.set_magic_keybinding(val)

        return False

    def get_use_command_keys(self) -> bool:
        return _strbool(self.get_config("Kupfer", "usecommandkeys"))

    def set_use_command_keys(self, enabled: bool) -> bool:
        return self.set_config("Kupfer", "usecommandkeys", enabled)

    def get_action_accelerator_modifer(self) -> str:
        return str(self.get_config("Kupfer", "action_accelerator_modifer"))

    def set_action_accelerator_modifier(self, value: str) -> bool:
        """
        Valid values are: 'alt', 'ctrl'
        """
        return self.set_config("Kupfer", "action_accelerator_modifer", value)

    def set_large_icon_size(self, size: str) -> bool:
        return self.set_config("Appearance", "icon_large_size", size)

    def set_small_icon_size(self, size: str) -> bool:
        return self.set_config("Appearance", "icon_small_size", size)

    def get_show_status_icon(self) -> bool:
        """Convenience: Show icon in notification area as bool (GTK)."""
        return _strbool(self.get_config("Kupfer", "showstatusicon"))

    def set_show_status_icon(self, enabled: bool) -> bool:
        """Set config value and return success"""
        return self.set_config("Kupfer", "showstatusicon", enabled)

    def get_show_status_icon_ai(self) -> bool:
        """Convenience: Show icon in notification area as bool (AyatanaAppIndicator3)"""
        return _strbool(self.get_config("Kupfer", "showstatusicon_ai"))

    def set_show_status_icon_ai(self, enabled: bool) -> bool:
        """Set config value and return success"""
        return self.set_config("Kupfer", "showstatusicon_ai", enabled)

    def get_directories(self, direct: bool = True) -> ty.Iterator[str]:
        """Yield directories to use as directory sources"""
        specialdirs = {
            k: getattr(GLib.UserDirectory, k)
            for k in dir(GLib.UserDirectory)
            if k.startswith("DIRECTORY_")
        }

        def get_special_dir(opt):
            if opt.startswith("USER_"):
                opt = opt[5:]
                if opt in specialdirs:
                    return GLib.get_user_special_dir(specialdirs[opt])

            return None

        level = "Direct" if direct else "Catalog"
        for direc in self.get_config("Directories", level):
            dpath = get_special_dir(direc)
            yield dpath or os.path.abspath(os.path.expanduser(direc))

    def set_directories(self, dirs: list[str]) -> bool:
        return self.set_config("Directories", "direct", dirs)

    def get_plugin_config(
        self,
        plugin: str,
        key: str,
        value_type: PlugConfigValueType = str,
        default: PlugConfigValue | None = None,
    ) -> PlugConfigValue | None:
        """Return setting @key for plugin names @plugin, try to coerce to
        type @value_type.
        Else return @default if does not exist, or can't be coerced
        """
        plug_section = f"plugin_{plugin}"
        if plug_section not in self._config:
            return default

        val = self._get_raw_config(plug_section, key)
        if val is None:
            return default

        if isinstance(value_type, type) and issubclass(
            value_type, ExtendedSetting
        ):
            val_obj: ExtendedSetting = value_type()
            val_obj.load(plugin, key, val)
            return val_obj

        value = default
        try:
            if value_type is bool:
                value = _strbool(val)
            elif value_type is list:
                assert isinstance(val, str)
                value = _strlist(val)
            elif isinstance(value_type, type):
                value = value_type(val)
            elif isinstance(value_type, ValueConverter):
                value = value_type(val, default=default)  # type: ignore

        except (ValueError, TypeError) as err:
            self.output_info(f"Error for load value {plug_section}.{key}", err)

        return value

    def get_plugin_config_bool(
        self, plugin: str, key: str, default: bool
    ) -> bool:
        res = self.get_plugin_config(plugin, key, _strbool, default)
        assert isinstance(res, bool)
        return res

    def set_plugin_config(
        self,
        plugin: str,
        key: str,
        value: PlugConfigValue,
        value_type: PlugConfigValueType = str,
    ) -> bool:
        """Try set @key for plugin names @plugin, coerce to @value_type first."""

        plug_section = f"plugin_{plugin}"
        self._emit_value_changed(plug_section, key, value)

        value_repr: int | str | float | None

        if isinstance(value, ExtendedSetting):
            value_repr = value.save(plugin, key)
        elif value_type is list:
            value_repr = str(value)
        elif value is None or isinstance(value, (str, float, int)):
            value_repr = value

        return self._set_raw_config(plug_section, key, value_repr)

    def get_accelerator(self, name: str) -> str | None:
        res = self.get_config("Keybindings", name)
        assert res is None or isinstance(res, str)
        return res

    def set_accelerator(self, name: str, key: str) -> bool:
        return self.set_config("Keybindings", name, key)

    def get_accelerators(self) -> dict[str, ty.Any]:
        return self._config["Keybindings"]

    def reset_keybindings(self) -> None:
        if key := self._get_from_defaults("Kupfer", "keybinding"):
            self.set_keybinding(str(key))

        if key := self._get_from_defaults("Kupfer", "magickeybinding"):
            self.set_magic_keybinding(str(key))

    def reset_accelerators(self) -> None:
        for key, value in self._get_from_defaults_section("Keybindings") or ():
            self.set_config("Keybindings", key, value)

    def get_preferred_tool(self, tool_id: str) -> str | None:
        """Get preferred ID for a @tool_id.

        Supported: 'terminal', 'editor'
        """
        res = self.get_config("Tools", tool_id)
        assert res is None or isinstance(res, str)
        return res

    def set_preferred_tool(self, tool_id: str, value: ty.Any) -> bool:
        return self.set_config("Tools", tool_id, value)

    ## Alternatives section
    ## Provide alternatives for each category
    ## for example the category "terminal"
    def get_valid_alternative_ids(
        self, category_key: str
    ) -> ty.Iterator[tuple[str, str]]:
        """Get a list of (id_, name) tuples for the given @category_key."""
        if category_key not in self._alternative_validators:
            return

        validator = self._alternative_validators[category_key]
        for id_, alternative in self._alternatives[category_key].items():
            if not validator or validator(alternative):
                name = alternative["name"]
                yield (id_, name)

    def get_all_alternatives(self, category_key: str) -> ty.Any:
        return self._alternatives[category_key]

    def get_preferred_alternative(
        self, category_key: str
    ) -> dict[str, ty.Any] | None:
        """Get preferred alternative dict for @category_key."""
        tool_id = self.get_preferred_tool(category_key)
        alternatives = self._alternatives[category_key]
        if alt := alternatives.get(tool_id):
            assert not alt or isinstance(alt, dict)
            return alt

        self.output_debug("Warning, no configuration for", category_key)
        return next(iter(alternatives.values()), None)

    def update_alternatives(
        self,
        category_key: str,
        alternatives: dict[str, ty.Any],
        validator: AltValidator | None,
    ) -> None:
        self._alternatives[category_key] = alternatives
        self._alternative_validators[category_key] = validator
        self.emit("alternatives-changed::" + category_key, category_key)

    def get_ellipsize_mode(self) -> Pango.EllipsizeMode:
        if self.get_config("Appearance", "ellipsize_mode") == "1":
            return Pango.EllipsizeMode.END

        return Pango.EllipsizeMode.MIDDLE


# Arguments: Section, Key, Value
# Detailed by 'section.key' in lowercase
GObject.signal_new(
    "value-changed",
    SettingsController,
    GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.DETAILED,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_PYOBJECT),
)

# Arguments: Plugin ID, Value
GObject.signal_new(
    "plugin-enabled-changed",
    SettingsController,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_INT),
)

# Arguments: Plugin ID, Value
GObject.signal_new(
    "plugin-toplevel-changed",
    SettingsController,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING, GObject.TYPE_INT),
)

# Arguments: Alternative-category
# Detailed by: category key, in lowercase
GObject.signal_new(
    "alternatives-changed",
    SettingsController,
    GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.DETAILED,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_STRING,),
)


# Get SettingsController instance
get_settings_controller = SettingsController.instance


def is_known_terminal_executable(exearg: str) -> bool:
    """Check is `exearg` is know terminal executable"""
    setctl = get_settings_controller()
    for term in setctl.get_all_alternatives("terminal").values():
        if exearg == term["argv"][0]:
            return True

    return False


def get_configured_terminal() -> dict[str, ty.Any] | None:
    """Return the configured Terminal object"""
    setctl = get_settings_controller()
    return setctl.get_preferred_alternative("terminal")



"""
This is a program of its own, that does not integrate with the
Kupfer process.
"""

import os
import typing as ty

import dbus
import gi
from dbus.mainloop.glib import DBusGMainLoop

gi.require_version("Gtk", "3.0")
gi.require_version("Keybinder", "3.0")

# pylint: disable=wrong-import-position
from gi.repository import Gtk, Keybinder  # noqa:E402

if ty.TYPE_CHECKING:
    from gettext import gettext as _


__all__ = ("main",)

SERV = "se.kaizer.kupfer"
OBJ = "/interface"
IFACE = "se.kaizer.kupfer.Listener"


def _get_all_keys() -> list[str]:
    try:
        bus = dbus.Bus()
        obj = bus.get_object(SERV, OBJ)
        iface = dbus.Interface(obj, IFACE)
        return iface.GetBoundKeys(byte_arrays=True)  # type: ignore
    except dbus.DBusException as exc:
        print(exc)
        print("Waiting for Kupfer to start..")
        return []


def _rebind_key(keystring: str, is_bound: bool) -> None:
    if is_bound:
        print("binding", keystring)
        Keybinder.bind(keystring, _relay_key, keystring)
    else:
        print("unbinding", keystring)
        Keybinder.unbind(keystring)


def _relay_key(key: str) -> None:
    print("Relaying", key)
    time = Keybinder.get_current_event_time()
    s_id = f"kupfer-{os.getpid()}_TIME{time}"
    bus = dbus.Bus()
    obj = bus.get_object(SERV, OBJ, introspect=False)
    iface = dbus.Interface(obj, IFACE)
    iface.RelayKeysFromDisplay(key, os.getenv("DISPLAY", ":0"), s_id)


def main() -> None:
    DBusGMainLoop(set_as_default=True)

    for key in _get_all_keys():
        _rebind_key(key, True)

    bus = dbus.Bus()
    bus.add_signal_receiver(
        _rebind_key, "BoundKeyChanged", dbus_interface=IFACE
    )
    sicon = Gtk.StatusIcon.new_from_icon_name("kupfer")
    display = os.getenv("DISPLAY", ":0")
    sicon.set_tooltip_text(
        _("Keyboard relay is active for display %s") % display
    )
    sicon.set_visible(True)
    try:
        Gtk.main()
    except KeyboardInterrupt as exc:
        raise SystemExit(0) from exc


if __name__ == "__main__":
    main()



import os

from kupfer import config
from kupfer.support import pretty


def _extend_path():
    # Inside a function to not leak variables to module namespace
    if not config.has_capability("CUSTOM_PLUGINS"):
        return

    # Add plugins in data directories
    __path__.extend(config.get_data_dirs("plugins"))

    # Add .zip files in plugins directories
    for directory in __path__.copy():
        try:
            with os.scandir(directory) as entries:
                if zipnames := [f for f in entries if f.name.endswith(".zip")]:
                    pretty.print_debug(__name__, "Adding", directory, zipnames)
                    __path__.extend(z.path for z in zipnames)

        except OSError as error:
            pretty.print_error(__name__, error)
            continue


_extend_path()



__kupfer_name__ = _("Shorten Links")
__kupfer_actions__ = ("ShortenLinks",)
__description__ = _("Create short aliases of long URLs")
__version__ = "2017.1"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>, US"

import typing as ty
import urllib.parse
import urllib.request

from kupfer.obj import Action, Leaf, OperationError, Source, UrlLeaf
from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class _ShortLinksService(Leaf):
    def __init__(self, name):
        Leaf.__init__(self, name, name)

    def get_icon_name(self):
        return "text-html"


class _GETService(_ShortLinksService, pretty.OutputMixin):
    """A unified shortener service working with GET requests"""

    host: str
    path: str
    url_key = "url"
    use_https = False

    def process(self, url: str) -> str:
        """Shorten @url or raise ValueError"""
        assert self.path
        query_string = urllib.parse.urlencode({self.url_key: url})
        try:
            pretty.print_debug(__name__, "Request", self.path + query_string)
            with urllib.request.urlopen(self.path + query_string) as resp:
                if resp.status != 200:  # noqa:PLR2004
                    raise ValueError(
                        f"Invalid response {resp.status}, {resp.reason}"
                    )

                result: bytes = resp.read()
                return result.strip().decode("utf-8")

        except OSError as exc:
            raise ValueError(exc) from exc

        return _("Error")


# NOTE: It's important that we use only sites that provide a stable API


class IsGd(_GETService):
    """
    Website: http://is.gd
    Reference: http://is.gd/apishorteningreference.php
    """

    path = "https://is.gd/create.php?format=simple&"

    def __init__(self):
        super().__init__("Is.gd")


class VGd(_GETService):
    """
    Website: http://v.gd
    Reference: http://v.gd/apishorteningreference.php

    Like is.gd, but v.gd always shows a preview page.
    """

    path = "https://v.gd/create.php?format=simple&"

    def __init__(self):
        super().__init__("V.gd")


class ShortenLinks(Action):
    """Shorten links with selected engine"""

    def __init__(self):
        super().__init__(_("Shorten With..."))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        try:
            result = iobj.process(leaf.object)
        except ValueError as exc:
            raise OperationError(str(exc)) from exc

        return UrlLeaf(result, result)

    def item_types(self):
        yield UrlLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield _ShortLinksService

    def object_source(self, for_item=None):
        return ServicesSource()

    def get_description(self):
        return __description__


class ServicesSource(Source):
    source_use_cache = False

    def __init__(self):
        super().__init__(_("Services"))

    def get_items(self):
        yield IsGd()
        yield VGd()

    def should_sort_lexically(self):
        return True

    def get_icon_name(self):
        return "applications-internet"



__kupfer_name__ = _("Top")
__kupfer_sources__ = ("TaskSource",)
__description__ = _("Show running tasks and allow sending signals to them")
__version__ = "2023-04-29"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

import operator
import os
import signal
import subprocess
import typing as ty
from pathlib import Path

from kupfer import plugin_support
from kupfer.obj import Action, Leaf, Source

if ty.TYPE_CHECKING:
    from gettext import gettext as _


__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "sort_order",
        "label": _("Sort Order"),
        "type": str,
        "value": _("Commandline"),
        "alternatives": [
            _("Commandline"),
            _("CPU usage (descending)"),
            _("Memory usage (descending)"),
        ],
    },
)


class Task(Leaf):
    def __init__(self, path, name, description=None):
        Leaf.__init__(self, path, name)
        self._description = description

    def get_description(self):
        return self._description

    def get_actions(self):
        yield SendSignal()

    def get_icon_name(self):
        return "applications-system"


class SendSignal(Action):
    def __init__(self):
        Action.__init__(self, _("Send Signal..."))

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        os.kill(leaf.object, iobj.object)

    def requires_object(self):
        return True

    def object_types(self):
        yield _Signal

    def object_source(self, for_item=None):
        return _SignalsSource()


class _Signal(Leaf):
    def get_description(self):
        return f"kill -{self.object} ..."


# get all signals from signal package
_SIGNALS = tuple(
    _Signal(getattr(signal, signame), signame[3:])
    for signame in sorted(dir(signal))
    if signame.startswith("SIG") and not signame.startswith("SIG_")
)


class _SignalsSource(Source):
    def __init__(self):
        Source.__init__(self, _("Signals"))

    def get_items(self):
        return _SIGNALS

    def provides(self):
        yield _Signal


class TaskSource(Source):
    task_update_interval_sec = 5
    source_use_cache = False
    source_prefer_sublevel = True

    def __init__(self, name=_("Running Tasks")):
        Source.__init__(self, name)
        self._version = 3

    def is_dynamic(self):
        return True

    def get_items(self):
        uid = os.getuid()
        with subprocess.Popen(
            ["top", "-b", "-n", "1", "-u", str(uid)],
            stdout=subprocess.PIPE,
            env={"LC_NUMERIC": "C"},
        ) as proc:
            if proc.stdout:
                content = proc.stdout.read()
            else:
                return

        processes = parse_top_output(content)
        sort_order = __kupfer_settings__["sort_order"]
        if sort_order == _("Memory usage (descending)"):
            processes = sorted(
                processes, key=operator.itemgetter(2), reverse=True
            )
        elif sort_order == _("Commandline"):
            processes = sorted(processes, key=operator.itemgetter(4))
        # default: by cpu

        fields = _(
            "pid: %(pid)s  cpu: %(cpu)g%%  mem: %(mem)g%%  time: %(time)s"
        )
        for pid, cpu, mem, ptime, cmd in processes:
            description = fields % {
                "pid": pid,
                "cpu": cpu,
                "mem": mem,
                "time": ptime,
            }
            yield Task(pid, cmd, description)

    def get_description(self):
        return _("Running tasks for current user")

    def get_icon_name(self):
        return "system"

    def provides(self):
        yield Task


def parse_top_output(out):
    """
    Yield tuples of (pid, cpu, mem, ptime, cmd)
    """
    # Assuming UTF-8 output
    fields_map = {}
    fields_count = 0
    header_read = False
    for line in out.split(b"\n"):
        line = line.decode("utf-8", "replace").strip()  # noqa: PLW2901
        if line == "":
            header_read = True
            continue

        if not header_read:
            continue

        if line.startswith("PID"):  # assume pid is first col
            fields_map = {name: pos for pos, name in enumerate(line.split())}
            fields_count = len(fields_map)
            continue  # skip header

        line_fields = line.split(None, fields_count - 1)
        pid = line_fields[0]
        cpu = line_fields[fields_map["%CPU"]]
        mem = line_fields[fields_map["%MEM"]]
        ptime = line_fields[fields_map["TIME+"]]
        cmd = line_fields[-1]

        # read command line
        proc_file = Path(f"/proc/{pid}/cmdline")
        if proc_file.is_file():  # also skip (finished) missing tasks
            cmd = proc_file.read_text(encoding="UTF=8").replace("\x00", " ")
            yield (int(pid), float(cpu), float(mem), ptime, cmd)



from __future__ import annotations

__kupfer_name__ = _("Archive Manager")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ()
__kupfer_actions__ = ("UnpackHere", "CreateArchive", "CreateArchiveIn")
__description__ = _("Use Archive Manager actions")
__version__ = "2023-05-15"
__author__ = "Ulrik, KB"

import os
import re
import shutil
import typing as ty

# since "path" is a very generic name, you often forget..
from os import path as os_path
from pathlib import Path

from kupfer import launch, plugin_support, runtimehelper
from kupfer.obj import Action, FileLeaf, Leaf, OperationError
from kupfer.obj.special import CommandNotAvailableLeaf
from kupfer.support import fileutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from kupfer.core import commandexec

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "archive_type",
        "label": _("Compressed archive type for 'Create Archive In'"),
        "type": str,
        "value": ".tar.gz",
        "alternatives": (
            ".7z",
            ".tar",
            ".tar.gz",
            ".tar.bz2",
            ".tar.xz",
            ".zip",
        ),
    },
    {
        "key": "tool",
        "label": _("Tool used for operations:"),
        "type": str,
        "value": "file-roller",
        "alternatives": (
            "File-roller",
            "7-zip",
            "7za",
        ),
    },
)

_EXTENSIONS_SET = (
    ".7z",
    ".Z",
    ".ace",
    ".alz",
    ".ar",
    ".arj",
    ".bz",
    ".bz2",
    ".cab",
    ".cbr",
    ".cbz",
    ".cpio",
    ".deb",
    ".dmg",
    ".ear",
    ".gz",
    ".iso",
    ".jar",
    ".lha",
    ".lz",
    ".lzh",
    ".lzma",
    ".lzo",
    ".rar",
    ".rpm",
    ".sit",
    ".snap",
    ".sqsh",
    ".tar",
    ".taz",
    ".tbz",
    ".tbz2",
    ".tgz",
    ".tlrz",
    ".tlz",
    ".tzo",
    ".tzst",
    ".war",
    ".xz",
    ".zip",
    ".zoo",
    ".zst",
)


def _tool_cmd_path(tool: str) -> str | None:
    if tool == "7-zip":
        return shutil.which("7z")

    if tool == "7za":
        return shutil.which("7za")

    return shutil.which("file-roller")


class UnpackHere(Action):
    def __init__(self) -> None:
        Action.__init__(self, _("Extract Here"))

    def activate(
        self,
        leaf: Leaf,
        iobj: Leaf | None = None,
        ctx: commandexec.ExecutionToken | None = None,
    ) -> Leaf | None:
        tool = __kupfer_settings__["tool"]
        if not _tool_cmd_path(tool):
            return CommandNotAvailableLeaf(__name__, __kupfer_name__, tool)

        filedir = str(Path(leaf.object).parent)
        if tool == "7-zip":
            launch.spawn_in_terminal(["7z", "x", leaf.object], filedir)
            return None

        if tool == "7za":
            launch.spawn_in_terminal(["7za", "x", leaf.object], filedir)
            return None

        launch.spawn_async_notify_as(
            "file-roller.desktop",
            ["file-roller", "--extract-here", leaf.object],
        )
        return None

    def valid_for_item(self, leaf):
        fname, ext = os.path.splitext(leaf.object)
        # check for standard extension or a multi-part rar extension
        ext = ext.lower()
        if ext in _EXTENSIONS_SET:
            return True

        # for multi-part archives must exists also file without number, ie:
        # Filename.rar Filename.r00 Filename.r01 etc
        # not sure we can allow decompress no-first archive
        if re.search(r".r\d+$", ext) is not None:
            return os_path.isfile(f"{fname}.rar")

        return False

    def item_types(self):
        yield FileLeaf

    def get_description(self):
        return _("Extract compressed archive")

    def get_icon_name(self):
        return "extract-archive"


def _make_archive(filepaths: ty.Iterable[str]) -> None:
    cmd = ["file-roller", "--add"]
    cmd.extend(filepaths)
    launch.spawn_async_notify_as("file-roller.desktop", cmd)


class CreateArchive(Action):
    def __init__(self):
        Action.__init__(self, _("Create Archive"))

    def activate(self, leaf, iobj=None, ctx=None):
        _make_archive((leaf.object,))

    def activate_multiple(self, objs):
        _make_archive(L.object for L in objs)

    def item_types(self):
        yield FileLeaf

    def get_description(self):
        return _("Create a compressed archive from folder")

    def get_icon_name(self):
        return "add-files-to-archive"

    def valid_for_item(self, leaf: Leaf) -> bool:
        """Allow to launch this action only with file-roller"""
        tool = __kupfer_settings__["tool"]
        return tool not in ("7-zip", "7za")


def _make_archive_in(
    ctx: commandexec.ExecutionToken,
    basename: str,
    dirpath: str,
    filepaths: ty.Iterable[str],
) -> str:
    archive_type = __kupfer_settings__["archive_type"]
    archive_path = fileutils.get_destpath_in_directory(
        dirpath, basename, archive_type
    )

    tool = __kupfer_settings__["tool"]
    if not _tool_cmd_path(tool):
        raise OperationError(f"Command {tool} not available")

    if tool in ("7-zip", "7za"):
        cmd = ["7za" if tool == "7za" else "7z", "a", archive_path]
        cmd.extend(filepaths)
        launch.spawn_in_terminal(cmd)

    else:
        cmd = ["file-roller", f"--add-to={archive_path}"]
        cmd.extend(filepaths)
        launch.spawn_async_notify_as("file-roller.desktop", cmd)

    runtimehelper.register_async_file_result(ctx, archive_path)
    return archive_path


class CreateArchiveIn(Action):
    def __init__(self):
        Action.__init__(self, _("Create Archive In..."))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        assert ctx
        dirpath = iobj.object
        basename = os_path.basename(leaf.object)
        _make_archive_in(ctx, basename, dirpath, (leaf.object,))

    def activate_multiple(self, objs, iobjs, ctx):
        # TRANS: Default filename (no extension) for 'Create Archive In...'
        basename = _("Archive")
        for iobj in iobjs:
            _make_archive_in(
                ctx, basename, iobj.object, (L.object for L in objs)
            )

    def item_types(self):
        yield FileLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield FileLeaf

    def valid_object(self, obj, for_item=None):
        return fileutils.is_directory_writable(obj.object)

    def get_description(self):
        return _("Create a compressed archive from folder")

    def get_icon_name(self):
        return "add-files-to-archive"



"""
virtualbox_vboxapi_support.py

Control VirtualBox via Python interface (vboxapi).
Only (?) Sun VirtualBox (no OSE).
"""

__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"
__version__ = "2018-09-07"

from contextlib import suppress

import vboxapi  # pylint: disable=import-error

from kupfer.plugin.virtualbox import constants as vbox_const
from kupfer.support import pretty

MONITORED_DIRS = None
IS_DYNAMIC = False
ICON = "VBox"
APP_ID = "virtualbox"

_ACTIONS = {
    vbox_const.VM_POWEROFF: lambda c: c.powerDown(),
    vbox_const.VM_ACPI_POWEROFF: lambda c: c.powerButton(),
    vbox_const.VM_PAUSE: lambda c: c.pause(),
    vbox_const.VM_REBOOT: lambda c: c.reset(),
    vbox_const.VM_RESUME: lambda c: c.resume(),
    vbox_const.VM_SAVE: lambda c: c.saveState(),
}


def _get_object_session():
    """get new session to vm"""
    pretty.print_debug(__name__, "_get_object_session start")
    vbox, session = None, None
    try:
        vbox = vboxapi.VirtualBoxManager(None, None)
        session = vbox.mgr.getSessionObject(vbox.vbox)
    except Exception as err:
        pretty.print_error(__name__, "virtualbox: get session error ", err)

    pretty.print_debug(__name__, "_get_object_session finished", vbox, session)
    return vbox, session


def _get_existing_session(vm_uuid):
    """get existing session by machine uuid"""
    pretty.print_debug(__name__, "_get_existing_session start")
    vbox, session = None, None
    try:
        vbox = vboxapi.VirtualBoxManager(None, None)
        session = vbox.mgr.getSessionObject(vbox.vbox)
        vbox.vbox.openExistingSession(session, vm_uuid)
    except Exception as err:
        pretty.print_error(
            __name__, "virtualbox: get session error", vm_uuid, err
        )

    pretty.print_debug(
        __name__, "_get_existing_session finished", vbox, session
    )
    return vbox, session


def get_machine_state(machine_id):
    """check vms state (on/off/paused)"""
    pretty.print_debug(__name__, "get_machine_state", machine_id)
    vbox, vbox_sess = _get_object_session()
    if vbox_sess is None:
        return vbox_const.VM_STATE_POWEROFF

    state = vbox_const.VM_STATE_POWERON
    try:
        vbox.vbox.openExistingSession(vbox_sess, machine_id)
        machine_state = vbox_sess.machine.state
        if machine_state == vbox.constants.MachineState_Paused:
            state = vbox_const.VM_STATE_PAUSED
        elif machine_state in (
            vbox.constants.MachineState_PoweredOff,
            vbox.constants.MachineState_Aborted,
            vbox.constants.MachineState_Starting,
        ):
            state = vbox_const.VM_STATE_POWEROFF
        elif machine_state == vbox.constants.MachineState_Saved:
            state = vbox_const.VM_STATE_SAVED

    except Exception as err:  # exception == machine is off (xpcom.Exception)
        # silently set state to off
        state = vbox_const.VM_STATE_POWEROFF
        pretty.print_debug(__name__, "get_machine_state error", err)

    with suppress(Exception):
        if vbox_sess.state == vbox.constants.SessionState_Open:
            vbox_sess.close()

    pretty.print_debug(__name__, "get_machine_state finish", machine_id, state)
    return state


def _machine_start(vm_uuid, mode):
    """Start virtual machine
    @param vm_uuid - uuid of virtual machine
    @param mode - mode: gui, headless
    """
    vbox, session = _get_object_session()
    if session:
        try:
            remote_sess = vbox.vbox.openRemoteSession(
                session, vm_uuid, mode, ""
            )
            remote_sess.waitForCompletion(-1)
        except Exception as err:
            pretty.print_error(
                __name__, "StartVM:", vm_uuid, "Mode ", mode, "error", err
            )

        with suppress(Exception):  # various errors (xpcom.Exception)
            if session.state == vbox.constants.SessionState_Open:
                session.close()


def _execute_machine_action(vm_uuid, action):
    """Start virtual machine
    @param vm_uuid - uuid of virtual machine
    @param action - function called on vbox session
    """
    vbox, session = _get_existing_session(vm_uuid)
    try:
        action(session.console)
    except Exception as err:
        pretty.print_error(
            __name__,
            "_execute_machine_action:",
            action,
            " vm:",
            vm_uuid,
            "error",
            err,
        )

    with suppress(Exception):  # various errors (xpcom.Exception)
        if session.state == vbox.constants.SessionState_Open:
            session.close()


def vm_action(action, vm_uuid):
    """change state of the virtual machine
    @param action - one of the const VM_*
    @param vm_uuid - virtual machine uuid
    """
    if action == vbox_const.VM_START_NORMAL:
        _machine_start(vm_uuid, "gui")
    elif action == vbox_const.VM_START_HEADLESS:
        _machine_start(vm_uuid, "headless")
    else:
        command = _ACTIONS[action]
        _execute_machine_action(vm_uuid, command)


def get_machines():
    """Get generator of items:
    (machine uuid, machine name, machine description)
    """
    pretty.print_debug(__name__, "get_machines start")
    vbox, vbox_sess = _get_object_session()
    if vbox_sess is None:
        return

    machines = vbox.getArray(vbox.vbox, "machines")
    for machine in machines:
        pretty.print_debug(
            __name__, "get_machines; found machine", machine.id, machine.name
        )
        description = machine.description or machine.OSTypeId
        yield (machine.id, machine.name, description)

    pretty.print_debug(__name__, "get_machines finished")


def unload():
    pass



from __future__ import annotations

__kupfer_name__ = _("Kupfer Actions")
__kupfer_sources__ = ("KupferActions",)
__description__ = _(
    "'Inverse' action executions - look for action and then select object to "
    "execute on."
)
__version__ = "2023-12-17"
__author__ = "KB"


import typing as ty

from kupfer import icons
from kupfer.core import settings, sources
from kupfer.obj import Source
from kupfer.obj.objects import ActionLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class KupferActions(Source):
    """Get all global actions that don't require additional object."""

    def __init__(self):
        Source.__init__(self, _("Kupfer Actions"))

    def _on_plugin_enabled(
        self, _setctl: ty.Any, _plugin_id: str, _enabled: bool | int
    ) -> None:
        self.mark_for_update()

    def initialize(self):
        setctl = settings.get_settings_controller()
        setctl.connect("plugin-enabled-changed", self._on_plugin_enabled)

    def get_items(self):
        # we can skip action_generators
        sctl = sources.get_source_controller()
        for actions in sctl.action_decorators.values():
            for action in actions:
                # skip actions that require extra object
                if action.requires_object():
                    continue

                yield ActionLeaf(action)

    def get_gicon(self):
        return icons.ComposedIcon("kupfer", "kupfer-execute")



"""
Common exceptions definition.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

import typing as ty

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__all__ = (
    "Error",
    "InvalidDataError",
    "InvalidLeafError",
    "NoDefaultApplicationError",
    "NoMultiError",
    "NotAvailableError",
    "OperationError",
)


class Error(Exception):
    pass


class InvalidDataError(Error):
    """The data is wrong for the given Leaf"""


class OperationError(Error):
    """Command execution experienced an error"""


class InvalidLeafError(OperationError):
    """The Leaf passed to an Action is invalid"""


class NotAvailableError(OperationError):
    """User-visible error message when an external tool is the wrong version."""

    def __init__(self, toolname: str):
        OperationError.__init__(
            self, _("%s does not support this operation") % toolname
        )


class NoMultiError(OperationError):
    def __init__(self):
        OperationError.__init__(
            self, _("Can not be used with multiple objects")
        )


class NoDefaultApplicationError(OperationError):
    pass



"""
Kupfer's Contacts API

Main definition and *constructor* classes.

Constructor classes such as EmailContact are used to conveniently construct
contacts with common traits. To *use* contacts, always use ContactLeaf, asking
for specific slots to be filled.
"""

from __future__ import annotations

import typing as ty

from kupfer import icons
from kupfer.obj.grouping import GroupingLeaf, Slots
from kupfer.support import validators

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from gi.repository import GdkPixbuf

    from kupfer.obj.base import Leaf

__author__ = (
    "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>, "
    "Karol Będkowski <karol.bedkowsk+gh@gmail.com>, "
    "Adi Sieker <adi@sieker.info>"
)

__all__ = (
    "AIMContact",
    "AddressContact",
    "ContactLeaf",
    "EmailContact",
    "GoogleTalkContact",
    "ICQContact",
    "IMContact",
    "JabberContact",
    "MSNContact",
    "PhoneContact",
    "QQContact",
    "SkypeContact",
    "YahooContact",
    "email_from_leaf",
)

EMAIL_KEY = "EMAIL"
NAME_KEY = "NAME"
PHONE_KEY = "PHONE"
ADDRESS_KEY = "ADDRESS"
LABEL_KEY = "LABEL"
JABBER_JID_KEY = "JID"
JABBER_STATUS_KEY = "JABBER_STATUS"
JABBER_RESOURCE_KEY = "JABBER_RESOURCE"
AIM_KEY = "AIM"
GOOGLE_TALK_KEY = "GOOGLE_TALK"
ICQ_KEY = "ICQ"
MSN_KEY = "MSN"
QQ_KEY = "QQ"
SKYPE_KEY = "SKYPE"
YAHOO_KEY = "YAHOO"


class ContactLeaf(GroupingLeaf):
    """Represent "human" contact leaf, like email, im address and so."""

    grouping_slots: tuple[str, ...] = ()

    def __init__(self, obj: ty.Any, name: str, image: ty.Any = None) -> None:
        self.image = image
        GroupingLeaf.__init__(self, obj, name)

    def get_icon_name(self) -> str:
        return "stock_person"

    def get_text_representation(self) -> str | None:
        return self.get_description()

    def get_thumbnail(
        self, width: int, height: int
    ) -> GdkPixbuf.Pixbuf | None:
        if self.image:
            return icons.get_pixbuf_from_data(self.image, width, height)

        return GroupingLeaf.get_thumbnail(self, width, height)


## E-mail convenience and constructors


def _get_email_from_url(url: str) -> str:
    """convert http://foo@bar.pl -> foo@bar.pl"""
    *_dummy, email = url.partition("://")
    return email or url


def email_from_leaf(leaf: Leaf) -> str | None:
    """Return an email address string if @leaf has a valid email address.

    @leaf may also be a TextLeaf or UrlLeaf.
    Return a false value if no valid email is found.
    """
    if isinstance(leaf, ContactLeaf):
        return leaf.get(EMAIL_KEY)  # type: ignore

    email = _get_email_from_url(str(leaf.object))
    return email if validators.is_valid_email(email) else None


class EmailContact(ContactLeaf):
    grouping_slots = (*ContactLeaf.grouping_slots, EMAIL_KEY)

    def __init__(self, email: str, name: str, image: ty.Any = None) -> None:
        slots = {EMAIL_KEY: email, NAME_KEY: name}
        ContactLeaf.__init__(self, slots, name, image)

    def repr_key(self) -> ty.Any:
        return self.object[EMAIL_KEY]

    def get_description(self) -> str:
        return self.object[EMAIL_KEY]  #  type:ignore

    def get_gicon(self) -> icons.GIcon:
        return icons.ComposedIconSmall(self.get_icon_name(), "stock_mail")


class IMContact(ContactLeaf):
    grouping_slots = (*ContactLeaf.grouping_slots, EMAIL_KEY)

    def __init__(
        self,
        im_id_kind: str,
        im_id: str,
        name: str,
        label: str | None = None,
        other_slots: Slots = None,
        image: ty.Any = None,
    ) -> None:
        self.im_id_kind = im_id_kind
        slots = {im_id_kind: im_id, NAME_KEY: name, LABEL_KEY: label}
        if other_slots:
            slots.update(other_slots)

        ContactLeaf.__init__(self, slots, name, image)
        self.kupfer_add_alias(im_id)

    def repr_key(self) -> ty.Any:
        return self.object[self.im_id_kind]

    def get_description(self) -> str:
        return (  # type:ignore
            self.object[LABEL_KEY] or self.object[self.im_id_kind]
        )


class JabberContact(IMContact):
    """Minimal class for all Jabber contacts."""

    grouping_slots = (*IMContact.grouping_slots, JABBER_JID_KEY)

    def __init__(
        self,
        jid: str,
        name: str,
        status: str | None = None,
        resource: str | None = None,
        slots: Slots = None,
        image: ty.Any = None,
    ) -> None:
        IMContact.__init__(
            self,
            JABBER_JID_KEY,
            jid,
            name or jid,
            other_slots=slots,
            image=image,
        )
        self._description: str = _("[%(status)s] %(userid)s/%(service)s") % {
            # TRANS: unknown user status
            "status": status or _("unknown"),
            "userid": jid,
            "service": resource or "",
        }

    def get_description(self) -> str:
        return self._description


class AIMContact(IMContact):
    grouping_slots = (*IMContact.grouping_slots, AIM_KEY)

    def __init__(
        self,
        id_: str,
        name: str,
        slots: Slots = None,
        image: ty.Any = None,
    ) -> None:
        IMContact.__init__(self, AIM_KEY, id_, name, _("Aim"), slots, image)


class GoogleTalkContact(IMContact):
    grouping_slots = (*IMContact.grouping_slots, GOOGLE_TALK_KEY)

    def __init__(
        self, id_: str, name: str, slots: Slots = None, image: ty.Any = None
    ) -> None:
        IMContact.__init__(
            self, GOOGLE_TALK_KEY, id_, name, _("Google Talk"), slots, image
        )


class ICQContact(IMContact):
    grouping_slots = (*IMContact.grouping_slots, ICQ_KEY)

    def __init__(
        self, id_: str, name: str, slots: Slots = None, image: ty.Any = None
    ) -> None:
        IMContact.__init__(self, ICQ_KEY, id_, name, _("ICQ"), slots, image)


class MSNContact(IMContact):
    grouping_slots = (*IMContact.grouping_slots, MSN_KEY)

    def __init__(
        self, id_: str, name: str, slots: Slots = None, image: ty.Any = None
    ) -> None:
        IMContact.__init__(self, MSN_KEY, id_, name, _("MSN"), slots, image)


class QQContact(IMContact):
    grouping_slots = (*IMContact.grouping_slots, QQ_KEY)

    def __init__(
        self, id_: str, name: str, slots: Slots = None, image: ty.Any = None
    ) -> None:
        IMContact.__init__(self, QQ_KEY, id_, name, _("QQ"), slots, image)


class YahooContact(IMContact):
    grouping_slots = (*IMContact.grouping_slots, YAHOO_KEY)

    def __init__(
        self, id_: str, name: str, slots: Slots = None, image: ty.Any = None
    ) -> None:
        IMContact.__init__(
            self, YAHOO_KEY, id_, name, _("Yahoo"), slots, image
        )


class SkypeContact(IMContact):
    grouping_slots = (*IMContact.grouping_slots, SKYPE_KEY)

    def __init__(
        self, id_: str, name: str, slots: Slots = None, image: ty.Any = None
    ) -> None:
        IMContact.__init__(
            self, SKYPE_KEY, id_, name, _("Skype"), slots, image
        )


class PhoneContact(ContactLeaf):
    grouping_slots = (*ContactLeaf.grouping_slots, EMAIL_KEY)

    def __init__(
        self,
        number: str,
        name: str,
        label: str,
        slots: Slots = None,
        image: ty.Any = None,
    ) -> None:
        pslots = {PHONE_KEY: number, NAME_KEY: name, LABEL_KEY: label}
        if slots:
            pslots.update(slots)

        ContactLeaf.__init__(self, pslots, name, image)

    def repr_key(self) -> ty.Any:
        return self.object[PHONE_KEY]

    def get_description(self) -> str:
        return f"{self.object[LABEL_KEY]}: {self.object[PHONE_KEY]}"


class AddressContact(ContactLeaf):
    grouping_slots = (*ContactLeaf.grouping_slots, EMAIL_KEY)

    def __init__(
        self,
        address: str,
        name: str,
        label: str,
        slots: Slots = None,
        image: ty.Any = None,
    ) -> None:
        aslots = {ADDRESS_KEY: address, NAME_KEY: name, LABEL_KEY: label}
        if slots:
            aslots.update(slots)

        ContactLeaf.__init__(self, aslots, name, image)

    def repr_key(self) -> ty.Any:
        return self.object[ADDRESS_KEY]



from __future__ import annotations

from gi.repository import Gdk, Gtk

from kupfer import config, version

__all__ = ["GetDataDialog", "ask_for_text"]


class GetDataDialog:
    def __init__(
        self,
        title: str,
        message: str = "",
        screen: Gdk.Screen | None = None,
        parent: Gtk.Window | None = None,
    ):
        """
        screen: Screen to use
        parent: Parent toplevel window
        """
        builder = Gtk.Builder()
        builder.set_translation_domain(version.PACKAGE_NAME)

        builder.add_from_file(config.get_data_file("getdata_dialog.ui"))
        builder.connect_signals(self)  # pylint: disable=no-member
        self.window = builder.get_object("dialoggetdata")
        self._box_fields = builder.get_object("box_fields")
        builder.get_object("label_title").set_text(title)
        if message:
            builder.get_object("label_message").set_text(message)
        else:
            builder.get_object("label_message").hide()

        if screen:
            self.window.set_screen(screen)

        if parent:
            self.window.set_transient_for(parent)

        self._text = None
        self._fields: dict[str, Gtk.Widget] = {}
        self._result: dict[str, str] | None = None

    def add_field(self, name: str, label: str, value: str) -> None:
        """Add text entry to dialog with `label` and `value`. `name` is internal
        name of value returned by dialog."""
        row = len(self._fields)

        wlabel = Gtk.Label()
        wlabel.set_alignment(0, 0.5)  # pylint: disable=no-member
        wlabel.set_markup(label)
        wlabel.set_selectable(False)
        self._box_fields.attach(wlabel, 0, row, 1, 1)

        widget = Gtk.Entry()
        widget.set_text(value)
        widget.set_hexpand(True)
        widget.set_activates_default(True)
        self._box_fields.attach(widget, 1, row, 1, 1)

        self._fields[name] = widget

    def run(self) -> dict[str, str] | None:
        """Run dialog, return key codes or None when user press cancel"""
        self._box_fields.show_all()
        self.window.set_keep_above(True)
        self.window.run()
        self.window.destroy()
        return self._result

    def on_buttonok_activate(self, _widget: Gtk.Widget) -> bool:
        self._result = {n: w.get_text() for n, w in self._fields.items()}
        self.window.hide()
        return True

    def on_buttonclose_activate(self, _widget: Gtk.Widget) -> bool:
        self.window.hide()
        return True


def ask_for_text(
    title: str,
    message: str,
    label: str = "",
    value: str = "",
    screen: Gdk.Screen | None = None,
    parent: Gtk.Window | None = None,
) -> str | None:
    dlg = GetDataDialog(
        title,
        message,
        screen=screen,
        parent=parent,
    )
    dlg.add_field("text", label, value)
    res = dlg.run()
    return res["text"] if res else None



import argparse
import gettext
import locale
import runpy
import sys
import typing as ty
from contextlib import suppress
from pathlib import Path

if ty.TYPE_CHECKING:
    from gettext import gettext as _

try:
    from kupfer import version_subst  # type:ignore
except ImportError:
    version_subst = None

__all__ = ("main",)


def _setup_locale_and_gettext() -> None:
    """Set up localization with gettext"""
    package_name = "kupfer"
    localedir = "./locale"
    for ldir in ("./locale", "/usr/local/share/locale/"):
        if Path(ldir).is_dir():
            localedir = ldir
            break

    if version_subst:
        package_name = version_subst.PACKAGE_NAME
        localedir = version_subst.LOCALEDIR

    # Install _() builtin for gettext; always returning unicode objects
    # also install ngettext()
    gettext.install(package_name, localedir=localedir, names=("ngettext",))
    # For Gtk.Builder, we need to call the C library gettext functions
    # As well as set the codeset to avoid locale-dependent translation
    # of the message catalog
    locale.bindtextdomain(package_name, localedir)
    locale.bind_textdomain_codeset(package_name, "UTF-8")
    # to load in current locale properly for sorting etc
    with suppress(locale.Error):
        locale.setlocale(locale.LC_ALL, "")


_setup_locale_and_gettext()


def _print(*args: ty.Any) -> None:
    enc = locale.getpreferredencoding(do_setlocale=False)
    sys.stdout.buffer.write(" ".join(args).encode(enc, "replace"))
    sys.stdout.buffer.write(b"\n")


def _make_help_text(
    program_options: list[tuple[str, str]], misc_options: list[tuple[str, str]]
) -> str:
    usage_string = _("Usage: kupfer [ OPTIONS | FILE ... ]")

    def format_options(opts):
        return "\n".join(f"  --{o:<15}  {h}" for o, h in opts)

    popts = format_options(program_options)
    mopts = format_options(misc_options)
    return f"{usage_string}\n\n{popts}\n\n{mopts}\n"


def _make_plugin_list() -> str:
    # require setup path and locales
    from kupfer.core import plugins  # pylint: disable=import-outside-toplevel

    plugin_header = _("Available plugins:")
    plugin_list = plugins.get_plugin_desc()
    return "\n".join((plugin_header, plugin_list))


def _get_options() -> argparse.Namespace:
    """Return a list of other application flags with --* prefix included."""

    from kupfer import version  # pylint: disable=import-outside-toplevel

    parser = argparse.ArgumentParser(
        prog=version.PROGRAM_NAME,
        description=version.SHORT_DESCRIPTION,
    )
    parser.add_argument(
        "--no-splash",
        action="store_true",
        help=_("do not present main interface on launch"),
    )
    parser.add_argument(
        "--list-plugins", action="store_true", help=_("list available plugins")
    )
    parser.add_argument(
        "--debug", action="store_true", help=_("enable debug info")
    )
    # TRANS: --exec-helper=HELPER is an internal command
    # TRANS: that executes a helper program that is part of kupfer
    parser.add_argument("--exec-helper", nargs=1, help=_("run plugin helper"))
    parser.add_argument(
        "--no-colors",
        action="store_true",
        help=_("do not use colored text in terminal"),
    )

    parser.add_argument(
        "--version",
        action="version",
        version=f"{version.PACKAGE_NAME}  {version.VERSION}",
    )

    # Fix sys.argv that can be None in exceptional cases
    if sys.argv[0] is None:
        sys.argv[0] = "kupfer"

    args = parser.parse_args()

    if args.list_plugins:
        _print(_gtkmain(_make_plugin_list))
        raise SystemExit

    if args.exec_helper:
        _exec_helper(args.exec_helper[0])
        raise SystemExit(1)

    return args


def _print_banner() -> None:
    # require setup path and locales
    from kupfer import version  # pylint: disable=import-outside-toplevel

    if not sys.stdout or not sys.stdout.isatty():
        return

    banner = _(
        "%(PROGRAM_NAME)s: %(SHORT_DESCRIPTION)s\n"
        "   %(COPYRIGHT)s\n"
        "   %(WEBSITE)s\n"
    ) % vars(version)
    _print(banner)


def _set_process_title() -> None:
    try:
        import setproctitle  # pylint: disable=import-outside-toplevel
    except ImportError:
        pass
    else:
        setproctitle.setproctitle("kupfer")


def _exec_helper(helpername: str) -> None:
    runpy.run_module(helpername, run_name="__main__", alter_sys=True)
    raise SystemExit


def _gtkmain(
    run_function: ty.Callable[..., ty.Any],
    *args: ty.Any,
    **kwargs: ty.Any,
) -> ty.Any:
    import gi  # pylint: disable=import-outside-toplevel

    gi.require_version("Gtk", "3.0")
    gi.require_version("Gdk", "3.0")
    with suppress(ValueError):
        gi.require_version("Wnck", "3.0")

    with suppress(ValueError):
        gi.require_version("AyatanaAppIndicator3", "0.1")

    return run_function(*args, **kwargs)


def _browser_start(quiet: bool) -> None:
    # pylint: disable=import-outside-toplevel
    from gi.repository import Gdk, GLib

    # program name; propagated to WM_NAME, WM_CLASS
    GLib.set_prgname("Kupfer")

    if not Gdk.Screen.get_default():  #pylint: disable=no-value-for-parameter
        print("No Screen Found, Exiting...", file=sys.stderr)
        sys.exit(1)

    from kupfer.ui import browser  # pylint: disable=import-outside-toplevel

    wctrl = browser.WindowController()
    wctrl.main(quiet=quiet)


def main() -> None:
    # parse commandline before importing UI
    cli_opts = _get_options()
    _print_banner()

    # pylint: disable=import-outside-toplevel
    from kupfer import version
    from kupfer.support import pretty

    if cli_opts.debug:
        pretty.DEBUG = True
        pretty.print_debug(
            __name__, "Version:", version.PACKAGE_NAME, version.VERSION
        )
        with suppress(ImportError):
            import debug

            debug.install()

    # enable colors only on terminal
    pretty.COLORS = sys.stdout.isatty() and not cli_opts.no_colors

    sys.excepthook = sys.__excepthook__
    _set_process_title()

    quiet = cli_opts.no_splash
    _gtkmain(_browser_start, quiet)



"""
Common objects for session_* plugins.
"""

import typing as ty

from kupfer import launch
from kupfer.obj import RunnableLeaf, Source
from kupfer.support import pretty
from kupfer.ui import uiutils

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__version__ = "2009-12-05"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"


def launch_argv_with_fallbacks(
    commands: list[str], print_error: bool = True
) -> bool:
    """Try the sequence of @commands with launch.spawn_async,
    and return with the first successful command.
    return False if no command is successful and log an error
    """
    for argv in commands:
        if ret := launch.spawn_async(argv):
            return ret

    pretty.print_error(__name__, "Unable to run command(s)", commands)
    return False


class CommandLeaf(RunnableLeaf):
    """The represented object of the CommandLeaf is a list of commandlines"""

    def run(self, ctx=None):
        launch_argv_with_fallbacks(self.object)


class Logout(CommandLeaf):
    """Log out from desktop"""

    def __init__(self, commands, name=None):
        super().__init__(commands, name or _("Log Out..."))

    def get_description(self):
        return _("Log out or change user")

    def get_icon_name(self):
        return "system-log-out"

    def run(self, ctx=None):
        if uiutils.confirm_dialog(_("Please confirm log out"), _("Log Out")):
            super().run(ctx)


class Shutdown(CommandLeaf):
    """Shutdown computer or reboot"""

    def __init__(self, commands, name=None):
        super().__init__(commands, name or _("Shut Down..."))

    def get_description(self):
        return _("Shut down, restart or suspend computer")

    def get_icon_name(self):
        return "system-shutdown"

    def run(self, ctx=None):
        if uiutils.confirm_dialog(
            _("Please confirm system shut down."), _("Shut down")
        ):
            super().run(ctx)


class LockScreen(CommandLeaf):
    """Lock screen"""

    def __init__(self, commands, name=None):
        super().__init__(commands, name or _("Lock Screen"))

    def get_description(self):
        return _("Enable screensaver and lock")

    def get_icon_name(self):
        return "system-lock-screen"


class CommonSource(Source):
    def is_dynamic(self):
        return True

    def get_icon_name(self):
        return "system-shutdown"

    def provides(self):
        yield RunnableLeaf



from __future__ import annotations

__kupfer_name__ = _("Recoll")
__kupfer_actions__ = ("Recoll", "OpenInRecoll")
__description__ = _("Search in Recoll full text search system")
__version__ = ""
__author__ = "KB"

import base64
import shutil
import subprocess
import typing as ty
from contextlib import suppress

from gi.repository import Gio

from kupfer import launch, plugin_support
from kupfer.obj import Action, FileLeaf, OperationError, Source, TextLeaf
from kupfer.obj.special import CommandNotAvailableLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _

plugin_support.check_command_available("recoll")


class RecollLeaf(FileLeaf):
    def __init__(self, fpath: str, title: str, mtype: str):
        super().__init__(fpath)
        self._title = title
        self._mtype = mtype
        if title != fpath:
            self.kupfer_add_alias(title)

    def get_description(self) -> str | None:
        return self._title

    def get_content_type(self) -> str | None:
        return self._mtype or super().get_content_type()


class Recoll(Action):
    rank_adjust: int = -5

    def __init__(self):
        Action.__init__(self, _("Search in Recoll"))

    def is_factory(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        return RecollQuerySource(leaf.object)

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Search for text in Recoll database")

    def get_icon_name(self):
        return "find"


class OpenInRecoll(Action):
    rank_adjust: int = -4

    def __init__(self):
        Action.__init__(self, _("Open in Recoll"))

    def activate(self, leaf, iobj=None, ctx=None):
        try:
            launch.spawn_async_raise(["recoll", "-q", str(leaf)])
        except launch.SpawnError as exc:
            raise OperationError(exc) from exc

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Open Recoll and search for text")

    def get_icon_name(self):
        return "find"


class RecollQuerySource(Source):
    def __init__(self, query):
        Source.__init__(self, name=_('Results for "%s"') % query)
        self._query = query.strip()
        self._max_items = 100

    def repr_key(self):
        return f"recoll_query:{self._query}"

    def get_items(self):
        recoll_cmd = shutil.which("recoll")
        if not recoll_cmd:
            yield CommandNotAvailableLeaf(__name__, __kupfer_name__, "recoll")
            return

        if not self._query:
            return

        command = [
            recoll_cmd,
            "-t",
            "-n",
            str(self._max_items),
            "-F",
            "url filename title mtype",
            "-S",
            "relevancyrating",
            "-D",
            self._query,
        ]

        with subprocess.Popen(command, stdout=subprocess.PIPE) as proc:
            out, _error = proc.communicate()
            for line in out.splitlines():
                if line.startswith(b"Recoll") or b"results" in line:
                    continue

                uri, filename, title, mtype, *_dummy = tuple(
                    base64.b64decode(v).decode() for v in line.split(b" ")
                )

                with suppress(OSError):
                    gfile = Gio.File.new_for_uri(uri)
                    fpath = gfile.get_path()
                    yield RecollLeaf(fpath, title or filename, mtype)

    def has_parent(self) -> bool:
        return False



from __future__ import annotations

__kupfer_name__ = _("Clipboards")
__kupfer_sources__ = ("ClipboardSource",)
__kupfer_actions__ = ("ClearClipboards",)
__description__ = _("Recent clipboards and clipboard proxy objects")
__version__ = "2023-03-22"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import typing as ty
from collections import deque
from contextlib import suppress
from pathlib import Path

from gi.repository import Gdk, Gio, Gtk

from kupfer import plugin_support
from kupfer.obj import (
    Action,
    FileLeaf,
    Leaf,
    Source,
    SourceLeaf,
    TextLeaf,
    UrlLeaf,
)
from kupfer.obj.compose import MultipleLeaf
from kupfer.support import validators

if ty.TYPE_CHECKING:
    from gettext import gettext as _
    from gettext import ngettext


__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "max",
        "label": _("Number of recent clipboards to remember"),
        "type": int,
        "value": 10,
    },
    {
        "key": "use_selection",
        "label": _("Include selected text in clipboard history"),
        "type": bool,
        "value": False,
    },
    {
        "key": "sync_selection",
        "label": _("Copy selected text to primary clipboard"),
        "type": bool,
        "value": False,
    },
)


class SelectedText(TextLeaf):
    qf_id = "selectedtext"

    def __init__(self, text):
        TextLeaf.__init__(self, text, _("Selected Text"))

    def __repr__(self):
        return f"<{__name__} {self.qf_id}>"


class SelectedFile(FileLeaf):
    qf_id = "selectedfile"

    def __init__(self, text):
        super().__init__(text, _("Selected File"))

    def __repr__(self):
        return f"<{__name__} {self.qf_id}>"


class SelectedUrl(UrlLeaf):
    qf_id = "selectedtext"

    def __init__(self, text):
        super().__init__(text, _("Selected URL"))

    def __repr__(self):
        return f"<{__name__} {self.qf_id}>"


def _new_selected_leaf(text: str | None) -> Leaf | None:
    if not text:
        return None

    if validators.is_url(text):
        return SelectedUrl(text)

    if validators.is_valid_file_path(text):
        path = Path(text).expanduser()
        if path.exists():
            return SelectedFile(path)

    return SelectedText(text)


class ClipboardText(TextLeaf):
    def get_description(self):
        numlines = self.object.count("\n") + 1
        desc = self.get_first_text_line(self.object)

        return ngettext(
            'Clipboard "%(desc)s"',
            'Clipboard with %(num)d lines "%(desc)s"',
            numlines,
        ) % {"num": numlines, "desc": desc}


class CurrentClipboardText(ClipboardText):
    qf_id = "clipboardtext"

    def __init__(self, text):
        ClipboardText.__init__(self, text, _("Clipboard Text"))

    def __repr__(self):
        return f"<{__name__} {self.qf_id}>"


class ClipboardUrl(UrlLeaf):
    def get_description(self):
        return _('Clipboard "%(desc)s"') % {"desc": self.object, "num": 1}


class ClipboardFile(FileLeaf):
    def get_description(self):
        return _('Clipboard "%(desc)s"') % {"desc": self.object, "num": 1}


class CurrentClipboardUrl(ClipboardUrl):
    qf_id = "clipboardtext"

    def __init__(self, text):
        super().__init__(text, _("Clipboard URL"))

    def __repr__(self):
        return f"<{__name__} {self.qf_id}>"


class CurrentClipboardFile(FileLeaf):
    "represents the *unique* current clipboard file"

    qf_id = "clipboardfile"

    def __init__(self, filepath):
        """@filepath is a filesystem byte string `str`"""
        FileLeaf.__init__(self, filepath, _("Clipboard File"))

    def __repr__(self):
        return f"<{__name__} {self.qf_id}>"


class CurrentClipboardFiles(MultipleLeaf):
    "represents the *unique* current clipboard if there are many files"

    qf_id = "clipboardfile"

    def __init__(self, paths):
        files = [FileLeaf(path) for path in paths]
        MultipleLeaf.__init__(self, files, _("Clipboard Files"))

    def __repr__(self):
        return f"<{__name__} {self.qf_id}>"


def _new_current_leaf(text: str | None) -> Leaf | None:
    if not text:
        return None

    if validators.is_url(text):
        return CurrentClipboardUrl(text)

    if validators.is_valid_file_path(text):
        path = Path(text).expanduser()
        with suppress(OSError):
            if path.exists():
                return CurrentClipboardFile(path)

    return CurrentClipboardText(text)


def _new_clipboard_leaf(text: str) -> Leaf:
    if validators.is_url(text):
        return ClipboardUrl(text, None)

    if validators.is_valid_file_path(text):
        with suppress(OSError, RuntimeError):
            path = Path(text).expanduser()
            if path.exists():
                return ClipboardFile(path)

    return ClipboardText(text)


class ClearClipboards(Action):
    def __init__(self):
        Action.__init__(self, _("Clear"))

    def activate(self, leaf, iobj=None, ctx=None):
        leaf.object.clear()

    def item_types(self):
        yield SourceLeaf

    def valid_for_item(self, leaf):
        return isinstance(leaf.object, ClipboardSource)

    def get_description(self):
        return _("Remove all recent clipboards")

    def get_icon_name(self):
        return "edit-clear"


class ClipboardSource(Source):
    source_scan_interval: int = 3600

    selected_text: str | None
    clipboard_uris: list[str]
    clipboard_text: str | None

    def __init__(self):
        Source.__init__(self, _("Clipboards"))
        self.clipboards: deque[str] = deque()

    # pylint: disable=attribute-defined-outside-init
    def initialize(self):
        clip = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        self._sig_id1 = clip.connect(
            "owner-change", self._on_clipboard_changed
        )
        clip = Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY)
        self._sig_id2 = clip.connect(
            "owner-change", self._on_clipboard_changed
        )
        self.clipboard_uris = []
        self.clipboard_text = None
        self.selected_text = None

    def finalize(self):
        clip = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clip.disconnect(self._sig_id1)
        clip = Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY)
        clip.disconnect(self._sig_id2)
        self._sig_id1 = None
        self._sig_id2 = None
        self.clipboard_uris = []
        self.clipboard_text = None
        self.selected_text = None
        self.mark_for_update()

    def _on_clipboard_changed(self, clip, event, *args):
        is_selection = event.selection == Gdk.SELECTION_PRIMARY
        clip.request_text(self._on_text_for_change, is_selection)
        if not is_selection:
            clip.request_uris(self._on_uris_for_change)

    def _on_text_for_change(self, clip, text, is_selection):
        if text is None:
            return

        max_len = __kupfer_settings__["max"]
        is_valid = bool(text and text.strip())
        is_sync_selection = (
            is_selection and __kupfer_settings__["sync_selection"]
        )

        if (
            not is_selection or __kupfer_settings__["use_selection"]
        ) and is_valid:
            self._add_to_history(text, is_selection)

        if is_sync_selection and is_valid:
            Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).set_text(text, -1)

        if is_selection:
            self.selected_text = text

        if not is_selection or is_sync_selection:
            self.clipboard_text = text

        self._prune_to_length(max_len)
        self.mark_for_update()

    def _on_uris_for_change(self, clip, uris):
        self.clipboard_uris = uris if uris is not None else []
        self.mark_for_update()

    def _add_to_history(self, cliptext, is_selection):
        if cliptext in self.clipboards:
            self.clipboards.remove(cliptext)
        # if the previous text is a prefix of the new selection, supersede it
        if (
            is_selection
            and self.clipboards
            and (
                cliptext.startswith(self.clipboards[-1])
                or cliptext.endswith(self.clipboards[-1])
            )
        ):
            self.clipboards.pop()

        self.clipboards.append(cliptext)

    def _prune_to_length(self, max_len):
        while len(self.clipboards) > max_len:
            self.clipboards.popleft()

    def get_items(self):
        # selected text
        if leaf := _new_selected_leaf(self.selected_text):
            yield leaf

        # produce the current clipboard files if any
        paths: list[str] = list(
            filter(
                None,
                (
                    Gio.File.new_for_uri(uri).get_path()
                    for uri in self.clipboard_uris
                ),
            )
        )

        if len(paths) == 1:
            yield CurrentClipboardFile(paths[0])
        elif len(paths) > 1:
            yield CurrentClipboardFiles(paths)

        # put out the current clipboard text
        if leaf := _new_current_leaf(self.clipboard_text):
            yield leaf

        # put out the clipboard history
        yield from map(_new_clipboard_leaf, reversed(self.clipboards))

    def get_description(self):
        return __description__

    def get_icon_name(self):
        return "edit-paste"

    def provides(self):
        yield TextLeaf
        yield FileLeaf
        yield MultipleLeaf

    def clear(self):
        self.clipboards.clear()
        self.mark_for_update()



__kupfer_name__ = _("Media Player Control")
__kupfer_sources__ = ("Players",)
__description__ = _("Playback control for media players")
__version__ = "2017.1"
__author__ = "US"

import typing as ty

import dbus

from kupfer import plugin_support
from kupfer.obj import OperationError, RunnableLeaf, Source
from kupfer.support import weaklib

if ty.TYPE_CHECKING:
    from gettext import gettext as _

plugin_support.check_dbus_connection()


_MPRIS_PREFIX = "org.mpris.MediaPlayer2."
_MPRIS_PLAYER = "org.mpris.MediaPlayer2.Player"
_MPRIS_OBJ = "/org/mpris/MediaPlayer2"


class Players(Source):
    source_use_cache = False

    def __init__(self):
        super().__init__(_("Media Player Control"))

    def initialize(self):
        bus = dbus.SessionBus()
        weaklib.dbus_signal_connect_weakly(
            bus,
            "NameOwnerChanged",
            self._name_owner_changed,
            dbus_interface="org.freedesktop.DBus",
        )

    def finalize(self):
        pass

    def _name_owner_changed(self, name, old, new):
        if name.startswith(_MPRIS_PREFIX):
            self.mark_for_update()

    def get_items(self):
        bus_names = []
        for bus_name in dbus.SessionBus().list_names():
            if bus_name.startswith(_MPRIS_PREFIX):
                name = bus_name[len(_MPRIS_PREFIX) :]
                yield PlayPause(bus_name, name)
                yield Next(bus_name, name)
                yield Previous(bus_name, name)
                yield Stop(bus_name, name)
                bus_names.append(bus_name)

        if bus_names:
            yield PauseAll(bus_names)

    def provides(self):
        yield RunnableLeaf

    def description(self):
        return __description__

    def get_icon_name(self):
        return "applications-multimedia"


def mpris_connection(bus_name, activate=False, operation_error=True):
    try:
        if obj := dbus.SessionBus().get_object(bus_name, _MPRIS_OBJ):
            return dbus.Interface(obj, _MPRIS_PLAYER)

    except dbus.exceptions.DBusException as err:
        raise OperationError(str(err)) from err

    return None


class MprisAction(RunnableLeaf):
    def __init__(self, bus_name, name):
        super().__init__(str(bus_name), name=name)

    @property
    def bus_name(self):
        return self.object

    def repr_key(self):
        return self.bus_name

    def run(self, ctx: ty.Any = None) -> None:
        raise NotImplementedError


class PlayPause(MprisAction):
    def __init__(self, bus_name, name):
        # TRANS: %s is a media player name
        super().__init__(bus_name, name=_("Play/Pause (%s)") % name)

    def run(self, ctx=None):
        mpris_connection(self.bus_name).PlayPause()

    def get_description(self):
        return _("Resume playback")

    def get_icon_name(self):
        return "media-playback-start"


def _reply_nop(*args):
    pass


class PauseAll(RunnableLeaf):
    def __init__(self, bus_names):
        super().__init__(name=_("Pause All"))
        self.bus_names = bus_names

    def run(self, ctx=None):
        for name in self.bus_names:
            mpris_connection(name).Pause(
                reply_handler=_reply_nop, error_handler=_reply_nop
            )

    def get_icon_name(self):
        return "media-playback-pause"


class Next(MprisAction):
    def __init__(self, bus_name, name):
        # TRANS: %s is a media player name
        super().__init__(bus_name, name=_("Next (%s)") % name)

    def run(self, ctx=None):
        mpris_connection(self.bus_name).Next()

    def get_description(self):
        return _("Skip to next track")

    def get_icon_name(self):
        return "media-skip-forward"


class Previous(MprisAction):
    def __init__(self, bus_name, name):
        # TRANS: %s is a media player name
        super().__init__(bus_name, name=_("Previous (%s)") % name)

    def run(self, ctx=None):
        mpris_connection(self.bus_name).Previous()

    def get_description(self):
        return _("Skip to previous track")

    def get_icon_name(self):
        return "media-skip-backward"


class Stop(MprisAction):
    def __init__(self, bus_name, name):
        # TRANS: %s is a media player name
        super().__init__(bus_name, name=_("Stop (%s)") % name)

    def run(self, ctx=None):
        mpris_connection(self.bus_name).Stop()

    def get_description(self):
        return _("Stop playback")

    def get_icon_name(self):
        return "media-playback-stop"
